# This file is part of the Blockchain Data Trading Simulator
#    https://gitlab.com/MatthiasLohr/bdtsim
#
# Copyright 2020 Matthias Lohr <mail@mlohr.com>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import copy
import itertools
from typing import Any, Callable, Dict, List, Optional, Tuple

from bdtsim.account import Account, seller, buyer
from bdtsim.funds_diff_collection import FundsDiffCollection
from bdtsim.protocol_path import Decision


class TransactionLogEntry(object):
    def __init__(self, account: Account, tx_dict: Dict[str, Any], tx_receipt: Dict[str, Any],
                 funds_diff_collection: FundsDiffCollection) -> None:
        self.account = account
        self.tx_dict = tx_dict
        self.tx_receipt = tx_receipt
        self.funds_diff_collection = funds_diff_collection


class TransactionLogList(object):
    class Aggregation(object):
        class Entry(object):
            def __init__(self, account: Account, tx_fees: int, tx_count: int, funds_diff: int) -> None:
                self.account = account
                self.tx_fees = tx_fees
                self.tx_count = tx_count
                self.funds_diff = funds_diff

            def __str__(self) -> str:
                return '%s: %d (%d transaction(s), value diff %d)' % (
                    self.account.name,
                    self.tx_fees,
                    self.tx_count,
                    self.funds_diff
                )

        def __init__(self, tx_list: 'TransactionLogList') -> None:
            self.entries: Dict[Account, TransactionLogList.Aggregation.Entry] = {}
            for tx in tx_list.tx_log_list:
                # work on log line
                entry = self.entries.get(tx.account)
                if entry is None:
                    self.entries.update({
                        tx.account: TransactionLogList.Aggregation.Entry(
                            tx.account,
                            int(tx.tx_receipt['gasUsed']),
                            1,
                            0
                        )
                    })
                else:
                    entry.tx_fees += int(tx.tx_receipt['gasUsed'])
                    entry.tx_count += 1
                # work on all fund diffs
                for account, funds_diff in tx.funds_diff_collection.items():
                    funds_diff_entry = self.entries.get(account)
                    if funds_diff_entry is None:
                        self.entries.update({
                            account: TransactionLogList.Aggregation.Entry(
                                account,
                                0,
                                0,
                                funds_diff
                            )
                        })
                    else:
                        funds_diff_entry.funds_diff += funds_diff

    def __init__(self) -> None:
        self.tx_log_list: List[TransactionLogEntry] = []
        self._aggregation: Optional[TransactionLogList.Aggregation] = None

    def __len__(self) -> int:
        return len(self.tx_log_list)

    def append(self, entry: TransactionLogEntry) -> None:
        return self.tx_log_list.append(entry)

    @property
    def aggregation(self) -> 'TransactionLogList.Aggregation':
        if self._aggregation is None:
            self._aggregation = TransactionLogList.Aggregation(self)
        return self._aggregation


class TransactionLogCollection(object):
    class Aggregation(object):
        class Entry(object):
            def __init__(self, account: Account, tx_fees_min: int, tx_fees_max: int, tx_fees_sum: int,
                         tx_count_min: int, tx_count_max: int, tx_count_sum: int, funds_diff_min: int,
                         funds_diff_max: int, list_count: int, ) -> None:
                self.account = account
                self.tx_fees_min = tx_fees_min
                self.tx_fees_max = tx_fees_max
                self.tx_fees_sum = tx_fees_sum
                self.tx_count_min = tx_count_min
                self.tx_count_max = tx_count_max
                self.tx_count_sum = tx_count_sum
                self.funds_diff_min = funds_diff_min
                self.funds_diff_max = funds_diff_max
                self.list_count = list_count

            def __str__(self) -> str:
                if self.tx_fees_min == self.tx_fees_max:
                    return '%s: %d (%d transaction(s), value diff %d)' % (
                        self.account.name,
                        self.tx_fees_min,
                        self.tx_count_min,
                        self.funds_diff_min
                    )
                else:
                    return '%s: %d/%d/%d (%d/%d/%d transaction(s), value diff %d/%d)' % (
                        self.account.name,
                        self.tx_fees_min, self.tx_fees_mean, self.tx_fees_max,
                        self.tx_count_min, self.tx_count_mean, self.tx_count_max,
                        self.funds_diff_min, self.funds_diff_max
                    )

            @property
            def tx_fees_mean(self) -> float:
                return self.tx_fees_sum / self.list_count

            @property
            def tx_count_mean(self) -> float:
                return self.tx_count_sum / self.list_count

        def __init__(self, tx_collection: 'TransactionLogCollection') -> None:
            self.entries: Dict[Account, TransactionLogCollection.Aggregation.Entry] = {}
            for tx_list in tx_collection.tx_log_lists:
                for tx_list_aggregation in tx_list.aggregation.entries.values():
                    entry = self.entries.get(tx_list_aggregation.account)
                    if entry is None:
                        self.entries.update({tx_list_aggregation.account: TransactionLogCollection.Aggregation.Entry(
                            tx_list_aggregation.account,
                            tx_list_aggregation.tx_fees,
                            tx_list_aggregation.tx_fees,
                            tx_list_aggregation.tx_fees,
                            tx_list_aggregation.tx_count,
                            tx_list_aggregation.tx_count,
                            tx_list_aggregation.tx_count,
                            tx_list_aggregation.funds_diff,
                            tx_list_aggregation.funds_diff,
                            1
                        )})
                    else:
                        entry.tx_fees_min = min(entry.tx_fees_min, tx_list_aggregation.tx_fees)
                        entry.tx_fees_max = max(entry.tx_fees_max, tx_list_aggregation.tx_fees)
                        entry.tx_fees_sum += tx_list_aggregation.tx_fees
                        entry.tx_count_min = min(entry.tx_count_min, tx_list_aggregation.tx_count)
                        entry.tx_count_max = max(entry.tx_count_max, tx_list_aggregation.tx_count)
                        entry.tx_count_sum += tx_list_aggregation.tx_count
                        entry.funds_diff_min = min(entry.funds_diff_min, tx_list_aggregation.funds_diff)
                        entry.funds_diff_max = max(entry.funds_diff_max, tx_list_aggregation.funds_diff)
                        entry.list_count += 1

        def __iadd__(self, other: 'TransactionLogCollection.Aggregation') -> 'TransactionLogCollection.Aggregation':
            if isinstance(other, TransactionLogCollection.Aggregation):
                for remote_entry in other.entries.values():
                    local_entry = self.entries.get(remote_entry.account)
                    if local_entry is None:
                        self.entries.update({remote_entry.account: copy.deepcopy(remote_entry)})
                    else:
                        local_entry.tx_fees_min += remote_entry.tx_fees_min
                        local_entry.tx_fees_max += remote_entry.tx_fees_max
                        local_entry.tx_fees_sum += remote_entry.tx_fees_sum
                        local_entry.tx_count_min += remote_entry.tx_count_min
                        local_entry.tx_count_max += remote_entry.tx_count_max
                        local_entry.tx_count_sum += remote_entry.tx_count_sum
                        local_entry.funds_diff_min += remote_entry.funds_diff_min
                        local_entry.funds_diff_max += remote_entry.funds_diff_max
                        local_entry.list_count += remote_entry.list_count
                return self
            else:
                return NotImplemented

    def __init__(self, tx_log_lists: Optional[List[TransactionLogList]] = None) -> None:
        self.tx_log_lists: List[TransactionLogList] = tx_log_lists or []
        self._aggregation: Optional[TransactionLogCollection.Aggregation] = None

    def __len__(self) -> int:
        return len(self.tx_log_lists)

    def append(self, entry: TransactionLogList) -> None:
        self.tx_log_lists.append(entry)

    @property
    def aggregation(self) -> 'TransactionLogCollection.Aggregation':
        if self._aggregation is None:
            self._aggregation = TransactionLogCollection.Aggregation(self)
        return self._aggregation


class ResultNode(object):
    def __init__(self, parent: Optional['ResultNode'] = None):
        self.parent = parent
        self.children: Dict[Decision, ResultNode] = {}
        self.tx_collection: TransactionLogCollection = TransactionLogCollection()

    def child(self, decision: Decision) -> 'ResultNode':
        child = self.children.get(decision)
        if child is None:
            child = ResultNode(self)
            self.children.update({decision: child})
        return child

    def all_accounts_completely_honest(self) -> bool:
        if self.parent is not None:
            incoming_decision = list(self.parent.children.keys())[list(self.parent.children.values()).index(self)]
            if not incoming_decision.is_honest():
                return False
            else:
                return self.parent.all_accounts_completely_honest()
        else:
            return True

    def account_completely_honest(self, account: Account) -> bool:
        if self.parent is not None:
            incoming_decision = list(self.parent.children.keys())[list(self.parent.children.values()).index(self)]
            if incoming_decision.account == account and not incoming_decision.is_honest():
                return False
            else:
                return self.parent.account_completely_honest(account)
        else:
            return True

    @property
    def final_nodes(self) -> List['ResultNode']:
        if len(self.children) > 0:
            return list(itertools.chain.from_iterable([c.final_nodes for c in self.children.values()]))
        else:
            return [self]

    @property
    def aggregation_summary(self) -> TransactionLogCollection.Aggregation:
        aggregation_summary = TransactionLogCollection.Aggregation(TransactionLogCollection())
        next_node: Optional[ResultNode] = self
        while next_node is not None:
            aggregation_summary += next_node.tx_collection.aggregation
            next_node = next_node.parent
        return aggregation_summary


class AggregationAttributeHelper(object):
    def __init__(self, account: Account, attribute: str) -> None:
        self._account = account
        self._attribute = attribute

    def __call__(self, aggregation: TransactionLogCollection.Aggregation) -> Any:
        return getattr(aggregation.entries.get(self._account), self._attribute)


class SimulationResult(object):
    def __init__(self) -> None:
        self.preparation_transactions = TransactionLogList()
        self.execution_result_root = ResultNode()
        self.cleanup_transactions = TransactionLogList()

    def get_important_execution_results(self) -> List[Tuple[str, TransactionLogCollection.Aggregation]]:
        all_honest_results = []
        seller_honest_results = []
        buyer_honest_results = []
        nobody_honest_results = []

        for node in self.execution_result_root.final_nodes:
            if node.all_accounts_completely_honest():
                all_honest_results.append(node.aggregation_summary)
            elif node.account_completely_honest(seller):
                seller_honest_results.append(node.aggregation_summary)
            elif node.account_completely_honest(buyer):
                buyer_honest_results.append(node.aggregation_summary)
            else:
                nobody_honest_results.append(node.aggregation_summary)

        important_results = []
        for honesty_str, results in [
            ('All Parties Honest', all_honest_results),
            ('Honest Seller, Cheating Buyer', seller_honest_results),
            ('Cheating Seller, Honest Buyer', buyer_honest_results),
            ('Cheating Seller, Cheating Buyer', buyer_honest_results)
        ]:
            for account in seller, buyer:
                tx_fees_max_result = self._apply_aggr_func(max, results, account, 'tx_fees_max')
                if tx_fees_max_result is not None:
                    important_results.append(('%s - %s max fees' % (honesty_str, account.name), tx_fees_max_result))

            profit_max_result = self._apply_aggr_func(max, results, seller, 'funds_diff_max')
            if profit_max_result is not None:
                important_results.append(('%s - Seller max profit' % honesty_str, profit_max_result))

            expenses_max_result = self._apply_aggr_func(min, results, buyer, 'funds_diff_min')
            if expenses_max_result is not None:
                important_results.append(('%s - Buyer max expenses' % honesty_str, expenses_max_result))

        return important_results

    @staticmethod
    def _apply_aggr_func(aggr_func: Callable[..., TransactionLogCollection.Aggregation],
                         results: List[TransactionLogCollection.Aggregation], account: Account,
                         attribute: str) -> Optional[TransactionLogCollection.Aggregation]:
        results_filtered = list(filter(lambda x: x.entries.get(account) is not None, results))
        if len(results_filtered) == 0:
            return None
        return aggr_func(results_filtered, key=AggregationAttributeHelper(account, attribute))



# This file is part of the Blockchain Data Trading Simulator
#    https://gitlab.com/MatthiasLohr/bdtsim
#
# Copyright 2020 Matthias Lohr <mail@mlohr.com>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Any, Dict, Optional

from .account import Account


class FundsDiffCollection(Dict[Account, int]):
    def __init__(self, init: Optional[Dict[Account, int]] = None) -> None:
        if init is None:
            super().__init__()
        else:
            super().__init__(init)

    def get(self, account: Account, default: int = 0) -> int:  # type: ignore
        return super().get(account, default)

    def __repr__(self) -> str:
        return '<%s.%s %s>' % (
            __name__,
            FundsDiffCollection.__name__,
            ', '.join(['%s: %d' % (account.name, diff) for account, diff in self.items()])
        )

    def __iadd__(self, other: Any) -> 'FundsDiffCollection':
        if isinstance(other, FundsDiffCollection):
            for account, other_amount in other.items():
                current_amount = self.get(account)
                if current_amount is None:
                    self.update({account: other_amount})
                else:
                    self.update({account: current_amount + other_amount})
            return self
        else:
            return NotImplemented

    def __add__(self, other: Any) -> 'FundsDiffCollection':
        if isinstance(other, FundsDiffCollection):
            result = FundsDiffCollection()
            result += self
            result += other
            return result
        else:
            return NotImplemented

    def __isub__(self, other: Any) -> 'FundsDiffCollection':
        if isinstance(other, FundsDiffCollection):
            for account, other_amount in other.items():
                current_amount = self.get(account)
                if current_amount is None:
                    self.update({account: -other_amount})
                else:
                    self.update({account: current_amount - other_amount})
            return self
        else:
            return NotImplemented

    def __sub__(self, other: Any) -> 'FundsDiffCollection':
        if isinstance(other, FundsDiffCollection):
            result = FundsDiffCollection()
            result += self
            result -= other
            return result
        else:
            return NotImplemented

    def __eq__(self, other: Any) -> bool:
        if isinstance(other, FundsDiffCollection):
            for a, b in (self, other), (other, self):
                for account in a.keys():
                    if a.get(account) != b.get(account):
                        return False
            return True
        else:
            return NotImplemented

    def __ne__(self, other: Any) -> bool:
        return not self.__eq__(other)



# This file is part of the Blockchain Data Trading Simulator
#    https://gitlab.com/MatthiasLohr/bdtsim
#
# Copyright 2020 Matthias Lohr <mail@mlohr.com>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
import time
from datetime import datetime
from typing import Any, Callable, Dict, Optional

from web3 import Web3
from web3.datastructures import AttributeDict
from web3.exceptions import TimeExhausted
from web3.gas_strategies.time_based import fast_gas_price_strategy
from web3.providers.base import BaseProvider
from web3.types import TxParams, Wei

from bdtsim.account import Account, buyer, seller, operator
from bdtsim.contract import SolidityContract
from bdtsim.funds_diff_collection import FundsDiffCollection


logger = logging.getLogger(__name__)


class Environment(object):
    def __init__(self, web3_provider: BaseProvider, chain_id: Optional[int] = None, gas_price: Optional[int] = None,
                 gas_price_strategy: Optional[Callable[[Web3, Optional[TxParams]], Wei]] = None) -> None:
        self._web3 = Web3(web3_provider)

        if chain_id is None:
            self._chain_id = self._web3.eth.chainId
            if self._chain_id is not None:
                logger.debug('Auto-detected chain id %d' % self._chain_id)
            else:
                logger.warning('Failed auto-detecting chain id')
        else:
            self._chain_id = chain_id

        self._gas_price = gas_price
        if gas_price_strategy is not None:
            self._web3.eth.setGasPriceStrategy(gas_price_strategy)
        else:
            self._web3.eth.setGasPriceStrategy(fast_gas_price_strategy)

        self._contract: Optional[SolidityContract] = None
        self._contract_address: Optional[str] = None

        self._transaction_callback: Optional[
            Callable[[Account, Dict[str, Any], Dict[str, Any], FundsDiffCollection], None]
        ] = None

    @property
    def chain_id(self) -> int:
        return self._chain_id

    @property
    def gas_price(self) -> Optional[int]:
        return self._gas_price

    def deploy_contract(self, account: Account, contract: SolidityContract, allow_failure: bool = False,
                        *args: Any, **kwargs: Any) -> None:
        web3_contract = self._web3.eth.contract(abi=contract.abi, bytecode=contract.bytecode)
        tx_receipt = self._send_transaction(
            account=account,
            factory=web3_contract.constructor(*args, **kwargs),
            allow_failure=allow_failure
        )
        self._contract_address = tx_receipt['contractAddress']
        logger.debug('New contract address: %s' % self._contract_address)
        self._contract = contract

    def send_contract_transaction(self, account: Account, method: str, *args: Any, value: int = 0,
                                  allow_failure: bool = False, **kwargs: Any) -> Any:
        if self._contract is None:
            raise RuntimeError('No contract available!')
        logger.debug('Preparing contract transaction %s(%s)' % (method, ', '.join([str(a) for a in [*args]])))
        web3_contract = self._web3.eth.contract(address=self._contract_address, abi=self._contract.abi)
        contract_method = getattr(web3_contract.functions, method)
        factory = contract_method(*args, **kwargs)
        self._send_transaction(
            account=account,
            factory=factory,
            value=value,
            allow_failure=allow_failure
        )
        # TODO implement contract return value

    def send_direct_transaction(self, account: Account, to: Account, value: int = 0,
                                allow_failure: bool = False) -> None:
        self._send_transaction(
            account=account,
            to=to,
            value=value,
            allow_failure=allow_failure
        )

    def _send_transaction(self, account: Account, factory: Optional[Any] = None, to: Optional[Account] = None,
                          value: int = 0, allow_failure: bool = False) -> AttributeDict[str, Any]:
        tx_dict = {
            'from': account.wallet_address,
            'nonce': self._web3.eth.getTransactionCount(account.wallet_address, 'pending'),
            'value': value,
            'chainId': self._chain_id,
            'gas': 4000000
        }
        if to is not None:
            tx_dict['to'] = to.wallet_address

        if factory is not None:
            tx_dict = factory.buildTransaction(tx_dict)
        else:
            tx_dict['gas'] = 21000

        if self._gas_price is not None:
            tx_dict['gasPrice'] = self._gas_price
        else:
            tx_dict['gasPrice'] = self._web3.eth.generateGasPrice(tx_dict)
        tx_signed = self._web3.eth.account.sign_transaction(tx_dict, private_key=account.wallet_private_key)

        # collect current account balances
        balances_before: Dict[Account, int] = {}
        for tmp_account in seller, buyer, operator:
            balances_before.update({tmp_account: self._web3.eth.getBalance(tmp_account.wallet_address, 'latest')})

        logger.debug('Submitting transaction %s...' % str(tx_dict))
        tx_hash = self._web3.eth.sendRawTransaction(tx_signed.rawTransaction)

        tx_receipt = None
        while tx_receipt is None:
            logger.debug('Waiting for transaction receipt (hash is %s)...' % str(tx_hash.hex()))
            try:
                tx_receipt = self._web3.eth.waitForTransactionReceipt(tx_hash, timeout=10)
            except TimeExhausted:
                pass

        if not allow_failure and not tx_receipt['status']:
            raise RuntimeError('Transaction execution not successful')

        logger.debug('Got receipt %s' % str(tx_receipt))

        # collect current account balances
        funds_diff_collection = FundsDiffCollection()
        for tmp_account in seller, buyer, operator:
            balance_after = self._web3.eth.getBalance(tmp_account.wallet_address, 'latest')
            balance_diff = balance_after - balances_before.get(tmp_account)
            if balance_diff != 0:
                funds_diff_collection += FundsDiffCollection({tmp_account: balance_diff})

        funds_diff_collection += FundsDiffCollection({account: tx_receipt['gasUsed'] * 1000000000})

        if self.transaction_callback is not None:
            self.transaction_callback(account, tx_dict, dict(tx_receipt), funds_diff_collection)
        return tx_receipt

    def wait(self, seconds: int) -> None:
        timeout = self._web3.eth.getBlock('latest').timestamp + seconds
        logger.debug('Waiting for %d seconds' % seconds)
        time.sleep(seconds)  # wake up earlier
        while True:
            current = self._web3.eth.getBlock('latest').timestamp
            if current > timeout:
                return
            else:
                logger.debug('Still waiting for timeout (current: %s, waiting for: %s)' % (
                    datetime.fromtimestamp(current).isoformat(),
                    datetime.fromtimestamp(timeout).isoformat()
                ))
                time.sleep(3)

    @property
    def web3(self) -> Web3:
        return self._web3

    @property
    def transaction_callback(self) -> Optional[
        Callable[[Account, Dict[str, Any], Dict[str, Any], FundsDiffCollection], None]
    ]:
        return self._transaction_callback

    @transaction_callback.setter
    def transaction_callback(self, callback: Optional[
        Callable[[Account, Dict[str, Any], Dict[str, Any], FundsDiffCollection], None]
    ]) -> None:
        self._transaction_callback = callback



# This file is part of the Blockchain Data Trading Simulator
#    https://gitlab.com/MatthiasLohr/bdtsim
#
# Copyright 2020 Matthias Lohr <mail@mlohr.com>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import BinaryIO, Optional

from .data_provider import DataProvider
from .data_provider_manager import DataProviderManager


class FileDataProvider(DataProvider):
    def __init__(self, filename: str) -> None:
        super(FileDataProvider, self).__init__()
        self._filename = filename
        self._fp: Optional[BinaryIO] = None

    @property
    def data_size(self) -> int:
        current = self.file_pointer.tell()  # save current position
        self.file_pointer.seek(0, 2)  # seek to the end of the file
        size = self.file_pointer.tell()  # save current position (size)
        self.file_pointer.seek(current, 0)  # move to previous position
        return size

    @property
    def file_pointer(self) -> BinaryIO:
        if self._fp is None:
            self._fp = open(self._filename, 'rb')
        return self._fp

    def __del__(self) -> None:
        if self._fp is not None:
            self._fp.close()
            self._fp = None


DataProviderManager.register('FileDataProvider', FileDataProvider)



# This file is part of the Blockchain Data Trading Simulator
#    https://gitlab.com/MatthiasLohr/bdtsim
#
# Copyright 2020 Matthias Lohr <mail@mlohr.com>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import unittest

from bdtsim.account import seller, buyer
from bdtsim.funds_diff_collection import FundsDiffCollection


class FundsDiffCollectionTest(unittest.TestCase):
    def test_eq(self):
        self.assertEqual(FundsDiffCollection({seller: 42}), FundsDiffCollection({seller: 42}))
        self.assertNotEqual(FundsDiffCollection({seller: 42}), FundsDiffCollection({buyer: 42}))
        self.assertNotEqual(FundsDiffCollection({seller: 42}), FundsDiffCollection({seller: 43}))

        self.assertEqual(FundsDiffCollection({seller: 42}), FundsDiffCollection({seller: 42, buyer: 0}))

    def test_iadd(self):
        fd = FundsDiffCollection({seller: 7})
        fd += FundsDiffCollection({seller: 6})
        self.assertEqual(FundsDiffCollection({seller: 13}), fd)

        fd = FundsDiffCollection({seller: 7})
        fd += FundsDiffCollection({buyer: 6})
        self.assertEqual(FundsDiffCollection({seller: 7, buyer: 6}), fd)

        fd = FundsDiffCollection({seller: 6, buyer: 7})
        fd += FundsDiffCollection({seller: 6})
        self.assertEqual(FundsDiffCollection({seller: 12, buyer: 7}), fd)
        fd += FundsDiffCollection({seller: 6, buyer: 7})
        self.assertEqual(FundsDiffCollection({seller: 18, buyer: 14}), fd)

    def test_add(self):
        fd = FundsDiffCollection({seller: 7})
        fd = fd + FundsDiffCollection({seller: 6})
        self.assertEqual(FundsDiffCollection({seller: 13}), fd)

        fd = FundsDiffCollection({seller: 7})
        fd = fd + FundsDiffCollection({buyer: 6})
        self.assertEqual(FundsDiffCollection({seller: 7, buyer: 6}), fd)

        fd = FundsDiffCollection({seller: 6, buyer: 7})
        fd = fd + FundsDiffCollection({seller: 6})
        self.assertEqual(FundsDiffCollection({seller: 12, buyer: 7}), fd)
        fd = fd + FundsDiffCollection({seller: 6, buyer: 7})
        self.assertEqual(FundsDiffCollection({seller: 18, buyer: 14}), fd)

    def test_isub(self):
        fd = FundsDiffCollection({seller: 7})
        fd -= FundsDiffCollection({seller: 6})
        self.assertEqual(FundsDiffCollection({seller: 1}), fd)

        fd = FundsDiffCollection({seller: 7})
        fd -= FundsDiffCollection({buyer: 6})
        self.assertEqual(FundsDiffCollection({seller: 7, buyer: -6}), fd)

        fd = FundsDiffCollection({seller: 6, buyer: 7})
        fd -= FundsDiffCollection({seller: 5})
        self.assertEqual(FundsDiffCollection({seller: 1, buyer: 7}), fd)
        fd -= FundsDiffCollection({seller: 6, buyer: 6})
        self.assertEqual(FundsDiffCollection({seller: -5, buyer: 1}), fd)

    def test_sub(self):
        fd = FundsDiffCollection({seller: 7})
        fd = fd - FundsDiffCollection({seller: 6})
        self.assertEqual(FundsDiffCollection({seller: 1}), fd)

        fd = FundsDiffCollection({seller: 7})
        fd = fd - FundsDiffCollection({buyer: 6})
        self.assertEqual(FundsDiffCollection({seller: 7, buyer: -6}), fd)

        fd = FundsDiffCollection({seller: 6, buyer: 7})
        fd = fd - FundsDiffCollection({seller: 5})
        self.assertEqual(FundsDiffCollection({seller: 1, buyer: 7}), fd)
        fd = fd - FundsDiffCollection({seller: 6, buyer: 6})
        self.assertEqual(FundsDiffCollection({seller: -5, buyer: 1}), fd)



# This file is part of the Blockchain Data Trading Simulator
#    https://gitlab.com/MatthiasLohr/bdtsim
#
# Copyright 2020 Matthias Lohr <mail@mlohr.com>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import subprocess
import unittest


class CommandTest(unittest.TestCase):
    def test_list_environments(self):
        p = subprocess.Popen(['bdtsim', 'list-environments'], stdout=subprocess.PIPE)
        out, err = p.communicate()
        self.assertEqual(p.returncode, 0)
        self.assertEqual(out.decode('utf-8').strip(), '\n'.join(['PyEVM', 'Web3HTTP', 'Web3Websocket', 'Web3IPC']))

    def test_list_protocols(self):
        p = subprocess.Popen(['bdtsim', 'list-protocols'], stdout=subprocess.PIPE)
        out, err = p.communicate()
        self.assertEqual(p.returncode, 0)
        self.assertEqual(out.decode('utf-8').strip(), '\n'.join([
            'FairSwap-FileSale',
            'SimplePayment',
            'SimplePayment-prepaid',
            'SimplePayment-prepaid-direct',
            'SimplePayment-postpaid',
            'SimplePayment-postpaid-direct'
        ]))

    def test_list_data_providers(self):
        p = subprocess.Popen(['bdtsim', 'list-data-providers'], stdout=subprocess.PIPE)
        out, err = p.communicate()
        self.assertEqual(p.returncode, 0)
        self.assertEqual(out.decode('utf-8').strip(), '\n'.join([
            'RandomDataProvider',
            'FileDataProvider'
        ]))

    def test_list_output_formats(self):
        p = subprocess.Popen(['bdtsim', 'list-output-formats'], stdout=subprocess.PIPE)
        out, err = p.communicate()
        self.assertEqual(p.returncode, 0)
        self.assertEqual(out.decode('utf-8').strip(), '\n'.join([
            'dot',
            'human-readable',
            'json',
            'yaml'
        ]))

    def test_run_simplepayment(self):
        p = subprocess.Popen(['bdtsim', 'run', 'SimplePayment', 'PyEVM'], stdout=subprocess.PIPE)
        out, err = p.communicate()
        self.assertEqual(p.returncode, 0)

    def test_run_simplepayment_prepaid(self):
        p = subprocess.Popen(['bdtsim', 'run', 'SimplePayment-prepaid', 'PyEVM'], stdout=subprocess.PIPE)
        out, err = p.communicate()
        self.assertEqual(p.returncode, 0)

    def test_run_simplepayment_prepaid_direct(self):
        p = subprocess.Popen(['bdtsim', 'run', 'SimplePayment-prepaid-direct', 'PyEVM'], stdout=subprocess.PIPE)
        out, err = p.communicate()
        self.assertEqual(p.returncode, 0)

    def test_run_simplepayment_postpaid(self):
        p = subprocess.Popen(['bdtsim', 'run', 'SimplePayment-postpaid', 'PyEVM'], stdout=subprocess.PIPE)
        out, err = p.communicate()
        self.assertEqual(p.returncode, 0)

    def test_run_simplepayment_postpaid_direct(self):
        p = subprocess.Popen(['bdtsim', 'run', 'SimplePayment-postpaid-direct', 'PyEVM'], stdout=subprocess.PIPE)
        out, err = p.communicate()
        self.assertEqual(p.returncode, 0)

    def test_run_fairswap_file_sale(self):
        p = subprocess.Popen(['bdtsim', 'run', 'FairSwap-FileSale', 'PyEVM', '-d', 'size', '256'],
                             stdout=subprocess.PIPE)
        out, err = p.communicate()
        self.assertEqual(p.returncode, 0)



# This file is part of the Blockchain Data Trading Simulator
#    https://gitlab.com/MatthiasLohr/bdtsim
#
# Copyright 2020 Matthias Lohr <mail@mlohr.com>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import math
from typing import List, Tuple, Type

from web3 import Web3

from bdtsim.util.xor import xor_crypt
from .merkle import MerkleTreeNode, MerkleTreeLeaf, MerkleTreeHashLeaf, from_leaves


B032 = b'\x00' * 32


class DecodingError(Exception):
    pass


class NodeDigestMismatchError(DecodingError):
    def __init__(self, in1: MerkleTreeLeaf, in2: MerkleTreeLeaf, out: MerkleTreeLeaf, index_in: int,
                 index_out: int, expected_digest: bytes, actual_digest: bytes) -> None:
        self.in1 = in1
        self.in2 = in2
        self.out = out
        self.index_in = index_in
        self.index_out = index_out
        self.expected_digest = expected_digest
        self.actual_digest = actual_digest


class LeafDigestMismatchError(NodeDigestMismatchError):
    pass


def crypt(value: bytes, index: int, key: bytes) -> bytes:
    return xor_crypt(value, Web3.solidityKeccak(['uint256', 'bytes32'], [index, key]))


def encode(root: MerkleTreeNode, key: bytes) -> MerkleTreeNode:
    leaves_enc = [crypt(leaf.data, index, key) for index, leaf in enumerate(root.leaves)]
    digests_enc = [crypt(digest, 2 * len(leaves_enc) + index, key) for index, digest in enumerate(root.digests_pack)]
    return from_leaves([MerkleTreeLeaf(x) for x in leaves_enc]
                       + [MerkleTreeHashLeaf(x) for x in digests_enc]
                       + [MerkleTreeHashLeaf(B032)])


def decode(root: MerkleTreeNode, key: bytes) -> Tuple[MerkleTreeNode, List[NodeDigestMismatchError]]:
    leaf_bytes_enc = root.leaves
    if not math.log2(len(leaf_bytes_enc)).is_integer():
        raise ValueError('Merkle Tree must have 2^x leaves')
    if leaf_bytes_enc[-1] != B032:
        raise ValueError('The provided Merkle Tree does not appear to be encoded')

    errors: List[NodeDigestMismatchError] = []
    digest_start_index = int(len(leaf_bytes_enc) / 2)
    node_index = 0
    digest_index = digest_start_index
    nodes: List[MerkleTreeNode] = [MerkleTreeLeaf(crypt(leaf_bytes_enc[i].data, i, key))
                                   for i in range(0, digest_start_index)]
    while len(nodes) > 1:
        nodes_new = []
        for i in range(0, len(nodes), 2):
            node = MerkleTreeNode(nodes[i], nodes[i + 1])
            expected_digest = crypt(leaf_bytes_enc[digest_index].data, digest_start_index + digest_index, key)

            if node_index < digest_start_index:
                error_type: Type[NodeDigestMismatchError] = LeafDigestMismatchError
                actual_digest = node.digest
            else:
                error_type = NodeDigestMismatchError
                actual_digest = Web3.solidityKeccak(['bytes32', 'bytes32'], [
                    crypt(leaf_bytes_enc[node_index].data, digest_start_index + node_index, key),
                    crypt(leaf_bytes_enc[node_index + 1].data, digest_start_index + node_index + 1, key)
                ])

            if expected_digest != actual_digest:
                errors.append(error_type(
                    in1=leaf_bytes_enc[node_index],
                    in2=leaf_bytes_enc[node_index + 1],
                    out=leaf_bytes_enc[digest_index],
                    index_in=node_index,
                    index_out=digest_index,
                    expected_digest=expected_digest,
                    actual_digest=actual_digest
                ))

            node_index += 2
            digest_index += 1
            nodes_new.append(node)

        nodes = nodes_new

    return nodes[0], errors



# This file is part of the Blockchain Data Trading Simulator
#    https://gitlab.com/MatthiasLohr/bdtsim
#
# Copyright 2020 Matthias Lohr <mail@mlohr.com>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import unittest

from bdtsim.util.filesize import FileSize


class UtilFileSizeTest(unittest.TestCase):
    def test_init_eq(self):
        self.assertRaises(ValueError, FileSize, 1.3)

        self.assertEqual(FileSize(123), FileSize(123))
        self.assertEqual(123, FileSize(123))
        self.assertEqual(1000000, FileSize('1M'))
        self.assertEqual(FileSize('1000k'), FileSize('1M'))
        self.assertEqual(1300, FileSize('1.3k'))

    def test_format_bytes(self):
        self.assertEqual('1k', FileSize(1000).human_readable())
        self.assertEqual('1M', FileSize(1000000).human_readable())
        self.assertEqual('1G', FileSize(1000000000).human_readable())
        self.assertEqual('150k', FileSize(150000).human_readable())



# This file is part of the Blockchain Data Trading Simulator
#    https://gitlab.com/MatthiasLohr/bdtsim
#
# Copyright 2020 Matthias Lohr <mail@mlohr.com>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json
from typing import Any

from bdtsim.account import Account
from bdtsim.protocol_path import Decision
from .output_format import OutputFormat
from .output_format_manager import OutputFormatManager
from .simulation_result import SimulationResult, TransactionLogEntry, TransactionLogList, TransactionLogCollection,\
    ResultNode


class JSONOutputFormat(OutputFormat):
    def __init__(self) -> None:
        super(JSONOutputFormat, self).__init__()

    def render(self, simulation_result: SimulationResult) -> None:
        print(json.dumps(simulation_result, cls=JSONEncoder))


class JSONEncoder(json.JSONEncoder):
    def default(self, o: Any) -> Any:
        if isinstance(o, Account):
            return {
                'name': o.name,
                'wallet_address': o.wallet_address
            }
        elif isinstance(o, bytes):
            return '0x' + o.hex()
        elif isinstance(o, Decision):
            return {
                'account': o.account,
                'description': o.description,
                'variants': o.variants,
                'honest_variants': o.honest_variants,
                'outcome': o.outcome,
                'timestamp': o.timestamp
            }
        elif isinstance(o, ResultNode):
            return {
                'aggregation_summary': o.aggregation_summary,
                'transaction_log_collection': o.tx_collection,
                'children': [{
                    'decision': decision,
                    'child': child
                } for decision, child in o.children.items()],
            }
        elif isinstance(o, SimulationResult):
            return {
                'important_execution_results': o.get_important_execution_results(),
                'preparation_transactions': o.preparation_transactions,
                'execution_result_root': o.execution_result_root,
                'cleanup_transaction': o.cleanup_transactions,
            }
        elif isinstance(o, TransactionLogCollection):
            return {
                'aggregation': o.aggregation,
                'transaction_log_lists': o.tx_log_lists
            }
        elif isinstance(o, TransactionLogCollection.Aggregation):
            return list(o.entries.values())
        elif isinstance(o, TransactionLogCollection.Aggregation.Entry):
            return {
                'account': o.account,
                'tx_fees_min': o.tx_fees_min,
                'tx_fees_max': o.tx_fees_max,
                'tx_fees_mean': o.tx_fees_mean,
                'tx_count_min': o.tx_count_min,
                'tx_count_max': o.tx_count_max,
                'tx_count_mean': o.tx_count_mean,
                'funds_diff_min': o.funds_diff_min,
                'funds_diff_max': o.funds_diff_max,
                'count': o.list_count
            }
        elif isinstance(o, TransactionLogEntry):
            return {
                'account': o.account,
                'transaction': o.tx_dict,
                'transaction_receipt': o.tx_receipt,
                'funds_diffs': [{
                    'account': account,
                    'funds_diff': diff
                } for account, diff in o.funds_diff_collection.items()]
            }
        elif isinstance(o, TransactionLogList):
            return {
                'aggregation': o.aggregation,
                'transactions': o.tx_log_list
            }
        elif isinstance(o, TransactionLogList.Aggregation):
            return list(o.entries.values())
        elif isinstance(o, TransactionLogList.Aggregation.Entry):
            return {
                'account': o.account,
                'tx_fees': o.tx_fees,
                'tx_count': o.tx_count,
                'funds_diff': o.funds_diff
            }
        else:
            return super(JSONEncoder, self).default(o)


OutputFormatManager.register('json', JSONOutputFormat)



# This file is part of the Blockchain Data Trading Simulator
#    https://gitlab.com/MatthiasLohr/bdtsim
#
# Copyright 2020 Matthias Lohr <mail@mlohr.com>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
import math
import random
from typing import Any, Optional, Union

from hexbytes.main import HexBytes
from jinja2 import Template
from web3 import Web3

from bdtsim.account import Account
from bdtsim.contract import SolidityContract
from bdtsim.data_provider import DataProvider
from bdtsim.environment import Environment
from bdtsim.protocol import Protocol, ProtocolManager, ProtocolInitializationError, ProtocolExecutionError
from bdtsim.protocol_path import ProtocolPath
from . import merkle, encoding


logger = logging.getLogger(__name__)


class FairSwap(Protocol):
    CONTRACT_TEMPLATE_FILE = 'FairFileSale.tpl.sol'
    CONTRACT_NAME = 'FileSale'

    def __init__(self, slices_count: int = 8, timeout: int = 600, *args: Any, **kwargs: Any) -> None:
        """
        Args:
            slices_count (int): number of slices in which the data is to be split
            timeout (int): Timeout in seconds before refund can be used
            *args (Any): Additional arguments
            **kwargs (Any): Additional keyword arguments
        """
        super(FairSwap, self).__init__(*args, **kwargs)

        self._slices_count = int(slices_count)
        if self._slices_count < 1:
            raise ProtocolInitializationError('n must be an int >= 1')

        if not math.log2(self._slices_count).is_integer():
            raise ProtocolInitializationError('slice_count must be a power of 2')
        self._merkle_tree_depth = int(math.log2(self._slices_count)) + 1

        self._timeout = int(timeout)

    def _get_contract(self, buyer: Account, price: int, slice_length: int, file_root_hash: bytes,
                      ciphertext_root_hash: bytes, key_hash: bytes) -> SolidityContract:
        """
        Args:
            buyer (Account): Account which has to pay and will receive the data
            price (int): Price of the data to be traded in Wei
            slice_length (int): length of a data slice, must be a multiple of 32
            file_root_hash (bytes): bytes32 root hash of the plain file's merkle tree
            ciphertext_root_hash (bytes): bytes32 root has of the encrypted file's merkle tree
            key_hash (bytes): bytes32 keccak hash of key to be used

        Returns:
            SolidityContract: Actual smart contract to be deployed with pre-filled values
        """
        with open(self.contract_path(__file__, FairSwap.CONTRACT_TEMPLATE_FILE)) as f:
            contract_template = Template(f.read())

        contract_code_rendered = contract_template.render(
            merkle_tree_depth=self._merkle_tree_depth,
            slice_length=slice_length,
            slices_count=self._slices_count,
            timeout=self._timeout,
            receiver=buyer.wallet_address,
            price=price,
            key_commitment=self.hex(key_hash),
            ciphertext_root_hash=self.hex(ciphertext_root_hash),
            file_root_hash=self.hex(file_root_hash)
        )

        return SolidityContract(FairSwap.CONTRACT_NAME, contract_code=contract_code_rendered)

    def execute(self, protocol_path: ProtocolPath, environment: Environment, data_provider: DataProvider,
                seller: Account, buyer: Account, price: int = 1000000000) -> None:
        """Execute a file transfer using the FairSwap protocol / FileSale contract.

        Args:
            protocol_path (ProtocolPath):
            environment (Environment):
            data_provider (DataProvider):
            seller (Account):
            buyer (Account):
            price (int):

        Returns:
            None:
        """
        # initial assignments and checks
        slice_length = int(data_provider.data_size / self._slices_count)
        if slice_length % 32 > 0:
            raise ProtocolExecutionError('A slice must have a multiple of 32 as length.'
                                         ' Configured for using %d slices,'
                                         ' therefore data must have a multiple of %d as length' % (
                                             self._slices_count,
                                             self._slices_count * 32
                                         ))

        plain_data = data_provider.file_pointer.read()
        plain_merkle_tree = merkle.from_bytes(plain_data, self._slices_count)
        key = self.generate_bytes(32, 1337)

        # === 1a: Seller: encrypt file for transmission
        encryption_decision = protocol_path.decide(
            seller, 'File Encryption/Transfer', ['expected', 'completely different', 'leaf forgery', 'hash forgery']
        )
        if encryption_decision == 'expected':
            encrypted_merkle_tree = encoding.encode(plain_merkle_tree, key)
        elif encryption_decision == 'completely different':
            encrypted_merkle_tree = encoding.encode(
                merkle.from_bytes(self.generate_bytes(data_provider.data_size), self._slices_count),
                key
            )
        elif encryption_decision == 'leaf forgery' or encryption_decision == 'hash forgery':
            # extract correct plain data
            plain_leaves_data = [leaf.data for leaf in plain_merkle_tree.leaves]
            plain_digests = plain_merkle_tree.digests_pack

            # forge first leaf
            plain_leaves_data[0] = b'\x00' * len(plain_leaves_data[0])

            if encryption_decision == 'hash forgery':
                plain_digests[0] = merkle.MerkleTreeNode(
                    merkle.MerkleTreeLeaf(plain_leaves_data[0]),
                    merkle.MerkleTreeLeaf(plain_leaves_data[1])
                ).digest

            encrypted_leaves_data = [encoding.crypt(data, i, key) for i, data in enumerate(plain_leaves_data)]
            encrypted_digests = [
                encoding.crypt(data, 2 * len(plain_leaves_data) + i, key) for i, data in enumerate(plain_digests)
            ]

            encrypted_merkle_tree = merkle.from_leaves([merkle.MerkleTreeLeaf(x) for x in encrypted_leaves_data]
                                                       + [merkle.MerkleTreeHashLeaf(x) for x in encrypted_digests]
                                                       + [merkle.MerkleTreeHashLeaf(encoding.B032)])
        else:
            raise NotImplementedError()

        # === 1b: Seller: deploy smart contract
        deployment_decision = protocol_path.decide(seller, 'Contract Deployment',
                                                   ['as expected by buyer', 'commitment to wrong key',
                                                    'unexpected file root hash', 'incorrect ciphertext root hash'])
        if deployment_decision == 'as expected by buyer':
            transfer_plain_root_hash = plain_merkle_tree.digest
            transfer_ciphertext_root_hash = encrypted_merkle_tree.digest
            transfer_key = key
        elif deployment_decision == 'commitment to wrong key':
            transfer_plain_root_hash = plain_merkle_tree.digest
            transfer_ciphertext_root_hash = encrypted_merkle_tree.digest
            transfer_key = self.generate_bytes(32, avoid=key)
        elif deployment_decision == 'unexpected file root hash':
            transfer_plain_root_hash = self.generate_bytes(32, avoid=plain_merkle_tree.digest)
            transfer_ciphertext_root_hash = encrypted_merkle_tree.digest
            transfer_key = key
        elif deployment_decision == 'incorrect ciphertext root hash':
            transfer_plain_root_hash = plain_merkle_tree.digest
            transfer_ciphertext_root_hash = self.generate_bytes(32, avoid=encrypted_merkle_tree.digest)
            transfer_key = key
        else:
            raise NotImplementedError()

        # actual contract deployment
        environment.deploy_contract(seller, self._get_contract(
            buyer=buyer,
            price=price,
            slice_length=int(slice_length / 32),
            file_root_hash=transfer_plain_root_hash,
            ciphertext_root_hash=transfer_ciphertext_root_hash,
            key_hash=Web3.solidityKeccak(['bytes32'], [transfer_key])
        ))

        # === 2: Buyer: Accept Contract/Pay Price ===
        if (plain_merkle_tree.digest == transfer_plain_root_hash
                and encrypted_merkle_tree.digest == transfer_ciphertext_root_hash):
            acceptance_decision = protocol_path.decide(buyer, 'Accept', variants=['yes', 'leave'])
            if acceptance_decision == 'yes':
                environment.send_contract_transaction(buyer, 'accept', value=price)
            elif acceptance_decision == 'leave':
                return
            else:
                raise NotImplementedError()
        else:
            return

        # === 3: Seller: Reveal key ===
        key_revelation_decision = protocol_path.decide(seller, 'Key Revelation', ['yes', 'leave'])
        if key_revelation_decision == 'yes':
            environment.send_contract_transaction(seller, 'revealKey', transfer_key)
        elif key_revelation_decision == 'leave':
            logger.debug('Seller: Leaving without Key Revelation')
            if protocol_path.decide(buyer, 'Refund', variants=['yes', 'no']) == 'yes':
                logger.debug('Buyer: Waiting for timeout to request refund')
                environment.wait(self._timeout + 1)
                environment.send_contract_transaction(buyer, 'refund')
            return
        else:
            raise NotImplementedError()

        # === 4: Buyer: Send Ok/Complain/Leave
        root_hash_enc = encrypted_merkle_tree.leaves[-2].data
        if encoding.crypt(root_hash_enc, 3 * self._slices_count - 2, transfer_key) != plain_merkle_tree.digest:
            if protocol_path.decide(buyer, 'complain about root', ['yes']) == 'yes':
                logger.debug('Buyer: Complaining about incorrect file root hash')
                root_hash_leaf = encrypted_merkle_tree.leaves[-2]
                proof = encrypted_merkle_tree.get_proof(root_hash_leaf)
                environment.send_contract_transaction(buyer, 'complainAboutRoot', root_hash_leaf.data, proof)
                return
        else:
            decrypted_merkle_tree, errors = encoding.decode(encrypted_merkle_tree, transfer_key)
            if len(errors) == 0:
                if protocol_path.decide(buyer, 'Confirm', variants=['yes', 'leave'],
                                        honest_variants=['yes', 'leave']) == 'yes':
                    environment.send_contract_transaction(buyer, 'noComplain')
                    return
            elif isinstance(errors[-1], encoding.LeafDigestMismatchError):
                if protocol_path.decide(buyer, 'Complain about Leaf', ['yes']) == 'yes':
                    error: encoding.NodeDigestMismatchError = errors[-1]
                    environment.send_contract_transaction(
                        buyer,
                        'complainAboutLeaf',
                        error.index_out,
                        error.index_in,
                        error.out.data,
                        error.in1.data_as_list(),
                        error.in2.data_as_list(),
                        encrypted_merkle_tree.get_proof(error.out),
                        encrypted_merkle_tree.get_proof(error.in1)
                    )
                    return
            else:
                if protocol_path.decide(buyer, 'Complain about Node', ['yes']) == 'yes':
                    error = errors[-1]
                    environment.send_contract_transaction(
                        buyer,
                        'complainAboutNode',
                        error.index_out,
                        error.index_in,
                        error.out.data,
                        error.in1.data,
                        error.in2.data,
                        encrypted_merkle_tree.get_proof(error.out),
                        encrypted_merkle_tree.get_proof(error.in1)
                    )
                    return

        # === 5: Seller: Finalize (when Buyer leaves in 4)
        if protocol_path.decide(seller, 'Request Payout', variants=['yes', 'no'],
                                honest_variants=['yes', 'no']) == 'yes':
            environment.wait(self._timeout + 1)
            environment.send_contract_transaction(seller, 'refund')

    @staticmethod
    def hex(value: Union[HexBytes, bytes, bytearray]) -> str:
        if isinstance(value, HexBytes):
            return value.hex()
        elif isinstance(value, bytes):
            return '0x%s' % value.hex()
        elif isinstance(value, bytearray):
            return '0x%s' % value.hex()
        else:
            raise ValueError('Type not supported')

    @staticmethod
    def generate_bytes(length: int = 32, seed: Optional[int] = None, avoid: Optional[bytes] = None) -> bytes:
        if seed is not None:
            random.seed(seed)
        tmp = avoid
        while tmp is None or tmp == avoid:
            tmp = bytes(bytearray(random.getrandbits(8) for _ in range(length)))
        return tmp


ProtocolManager.register('FairSwap-FileSale', FairSwap)



# This file is part of the Blockchain Data Trading Simulator
#    https://gitlab.com/MatthiasLohr/bdtsim
#
# Copyright 2020 Matthias Lohr <mail@mlohr.com>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import List, Optional
from uuid import uuid4

from graphviz import Digraph  # type: ignore

from bdtsim.protocol_path import Decision
from bdtsim.util.types import to_bool
from .output_format import OutputFormat
from .output_format_manager import OutputFormatManager
from .simulation_result import SimulationResult, ResultNode, TransactionLogList, TransactionLogCollection


class GraphvizDotOutputFormat(OutputFormat):
    COLOR_HONEST = '#00CC00'
    COLOR_CHEATING = '#FF0000'

    def __init__(self, output_filename: Optional[str] = None, view: bool = False, cleanup: bool = False,
                 output_format: str = 'pdf', graphviz_renderer: Optional[str] = None,
                 graphviz_formatter: Optional[str] = None, show_edges_details: Optional[bool] = True) -> None:
        """Create a [dot graph](https://www.graphviz.org/) for simulation result presentation.

        Args:
            output_filename (str): If provided, save dot graph to this file instead of printing to stdout
            view (bool): Open the rendered result with the default application (defaults to False).
            cleanup (bool): Delete the source file after rendering (defaults to False).
            output_format: The output format used for rendering (``'pdf'``, ``'png'``, etc., defaults to ``'pdf'``).
            graphviz_renderer: The output renderer used for rendering (``'cairo'``, ``'gd'``, ...).
            graphviz_formatter: The output formatter used for rendering (``'cairo'``, ``'gd'``, ...).
        """
        super(GraphvizDotOutputFormat, self).__init__()
        self._output_filename = output_filename
        self._view = view
        self._cleanup = cleanup
        self._output_format = output_format
        self._graphviz_renderer = graphviz_renderer
        self._graphviz_formatter = graphviz_formatter
        self._show_edges_details = to_bool(show_edges_details)

    def render(self, simulation_result: SimulationResult) -> None:
        graph = self._generate_graph(simulation_result)

        if self._output_filename is not None:
            graph.render(
                filename=self._output_filename,
                view=self._view,
                cleanup=self._cleanup,
                format=self._output_format,
                renderer=self._graphviz_renderer,
                formatter=self._graphviz_formatter
            )
        else:
            print(graph.source)

    def _generate_graph(self, simulation_result: SimulationResult) -> Digraph:
        graph = Digraph(
            graph_attr=[
                ('rankdir', 'LR'),
                ('fontname', 'helvetica')
            ],
            node_attr=[
                ('fontname', 'helvetica')
            ],
            edge_attr=[
                ('fontname', 'helvetica')
            ]
        )

        self._add_preparation_node(graph, simulation_result.preparation_transactions)
        self._add_cleanup_node(graph, simulation_result.cleanup_transactions)

        start_node_uuid = self._uuid()
        self._add_start_node(graph, start_node_uuid)
        self._walk_nodes(graph, simulation_result.execution_result_root, start_node_uuid, None)

        return graph

    def _add_preparation_node(self, graph: Digraph, tx_list: TransactionLogList) -> None:
        self._add_transaction_summaring_node(graph, 'Preparation', tx_list)

    def _add_cleanup_node(self, graph: Digraph, tx_list: TransactionLogList) -> None:
        self._add_transaction_summaring_node(graph, 'Cleanup', tx_list)

    @staticmethod
    def _add_transaction_summaring_node(graph: Digraph, title: str, tx_list: TransactionLogList) -> None:
        if len(tx_list) == 0:
            return

        label = title

        for entry in tx_list.aggregation.entries.values():
            label += '\n%s: %d (%d)' % (entry.account.name, entry.tx_fees, entry.tx_count)

        graph.node(title, label, shape='box')

    @staticmethod
    def _add_start_node(graph: Digraph, uuid: str) -> None:
        graph.node(uuid, label='<<b>Start</b>>')

    @staticmethod
    def _add_decision_node(graph: Digraph, uuid: str, label: str) -> None:
        graph.node(uuid, label=label, shape='diamond')

    def _add_start_edge(self, graph: Digraph, parent_uuid: str, child_uuid: str,
                        tx_collection: TransactionLogCollection) -> None:
        if self._show_edges_details:
            label = '<%s>' % ''.join(
                ['%s<br />' % str(label_line) for label_line in self._get_label_lines_for_tx_collection(tx_collection)]
            )
        else:
            label = ''

        graph.edge(
            tail_name=parent_uuid,
            head_name=child_uuid,
            label=label
        )

    def _add_decision_edge(self, graph: Digraph, parent_uuid: str, child_uuid: str, decision: Decision,
                           tx_collection: TransactionLogCollection) -> None:
        if self._show_edges_details:
            label_lines = self._get_label_lines_for_tx_collection(tx_collection)
            label = '<<b>%s</b><br />%s>' % (decision.outcome, ''.join(
                ['%s<br />' % str(line) for line in label_lines]
            ))
        else:
            label = '<<b>%s</b>>' % decision.outcome

        graph.edge(
            tail_name=parent_uuid,
            head_name=child_uuid,
            label=label,
            color=self._color_by_honesty(decision.is_honest())
        )

    @staticmethod
    def _get_label_lines_for_tx_collection(tx_collection: TransactionLogCollection) -> List[str]:
        return [str(entry) for entry in tx_collection.aggregation.entries.values()]

    def _add_end_node(self, graph: Digraph, uuid: str, node: ResultNode) -> None:
        label_lines = []
        for entry in node.aggregation_summary.entries.values():
            label_lines.append('<font color="%s">%s</font>' % (
                self._color_by_honesty(node.account_completely_honest(entry.account)),
                str(entry)
            ))

        graph.node(
            name=uuid,
            label='<<b>Total</b><br align="left" />%s<br align="left" />>' % '<br align="left" />'.join(label_lines),
            shape='box'
        )

    def _walk_nodes(self, graph: Digraph, current_node: ResultNode, parent_uuid: str,
                    incoming_decision: Optional[Decision]) -> None:
        current_node_uuid = self._uuid()

        if len(current_node.children) > 0:
            # Decision Node
            first_decision = list(current_node.children.keys())[0]
            current_node_label = '<%s<br /><b>%s</b>>' % (first_decision.account.name, first_decision.description)
            self._add_decision_node(graph, current_node_uuid, current_node_label)
        else:
            self._add_end_node(graph, current_node_uuid, current_node)

        if incoming_decision is not None:
            self._add_decision_edge(graph, parent_uuid, current_node_uuid, incoming_decision,
                                    current_node.tx_collection)
        else:
            self._add_start_edge(graph, parent_uuid, current_node_uuid, current_node.tx_collection)

        for decision, child_node in current_node.children.items():
            self._walk_nodes(graph, child_node, current_node_uuid, decision)

    @staticmethod
    def _uuid() -> str:
        return str(uuid4())

    @staticmethod
    def _color_by_honesty(honest: bool) -> str:
        return GraphvizDotOutputFormat.COLOR_HONEST if honest else GraphvizDotOutputFormat.COLOR_CHEATING


OutputFormatManager.register('dot', GraphvizDotOutputFormat)



# This file is part of the Blockchain Data Trading Simulator
#    https://gitlab.com/MatthiasLohr/bdtsim
#
# Copyright 2020 Matthias Lohr <mail@mlohr.com>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import unittest
from bdtsim.account import seller, buyer
from bdtsim.protocol_path import Decision, ProtocolPath


class DecisionTest(unittest.TestCase):
    def test_init(self):
        # should pass:
        Decision(seller, 'yes', ['yes', 'no'])

        self.assertRaises(ValueError, Decision, seller, 'no', ['yes'])

    def test_equals(self):
        self.assertEqual('yes', Decision(seller, 'yes', ['yes', 'no', 'maybe']))
        self.assertEqual(Decision(seller, 'yes', ['yes', 'no', 'maybe']), 'yes')
        self.assertEqual(Decision(seller, 'yes', ['yes', 'no', 'maybe']),
                         Decision(seller, 'yes', ['yes', 'no', 'maybe']))

        self.assertNotEqual(Decision(seller, 'yes', ['yes', 'no', 'maybe']),
                            Decision(buyer, 'yes', ['yes', 'no', 'maybe']))
        self.assertNotEqual(Decision(seller, 'yes', ['yes', 'no', 'maybe']),
                            Decision(seller, 'no', ['yes', 'no', 'maybe']))
        self.assertNotEqual(Decision(seller, 'yes', ['yes', 'no', 'maybe']),
                            Decision(seller, 'yes', ['yes', 'no', 'maybe', 'never']))


class ProtocolPathTest(unittest.TestCase):
    def test_get_alternatives2(self):
        path = ProtocolPath()
        path.decide(seller, '', ['yes', 'no'])
        self.assertEqual([
            ProtocolPath([Decision(seller, 'no', ['yes', 'no'])])
        ], path.get_alternatives())

        path.decide(buyer, '', ['yes', 'no'])
        self.assertEqual([
            ProtocolPath([Decision(seller, 'no', ['yes', 'no'])]),
            ProtocolPath([Decision(seller, 'yes', ['yes', 'no']), Decision(buyer, 'no', ['yes', 'no'])])
        ], path.get_alternatives())

    def test_get_alternatives3(self):
        path = ProtocolPath()
        path.decide(seller, '', ['yes', 'no', 'maybe'])
        self.assertEqual([
            ProtocolPath([Decision(seller, 'no', ['yes', 'no', 'maybe'])]),
            ProtocolPath([Decision(seller, 'maybe', ['yes', 'no', 'maybe'])])
        ], path.get_alternatives())

        path.decide(buyer, '', ['yes', 'no', 'maybe'])
        self.assertEqual([
            ProtocolPath([Decision(seller, 'no', ['yes', 'no', 'maybe'])]),
            ProtocolPath([Decision(seller, 'maybe', ['yes', 'no', 'maybe'])]),
            ProtocolPath([Decision(seller, 'yes', ['yes', 'no', 'maybe']),
                          Decision(buyer, 'no', ['yes', 'no', 'maybe'])]),
            ProtocolPath([Decision(seller, 'yes', ['yes', 'no', 'maybe']),
                          Decision(buyer, 'maybe', ['yes', 'no', 'maybe'])])
        ], path.get_alternatives())



# This file is part of the Blockchain Data Trading Simulator
#    https://gitlab.com/MatthiasLohr/bdtsim
#
# Copyright 2020 Matthias Lohr <mail@mlohr.com>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import unittest

from bdtsim.account import Account, buyer, operator, seller


class AccountTest(unittest.TestCase):
    def test_instantiation(self):
        self.assertRaises(ValueError, Account, 'john')

        self.assertRaises(ValueError, Account, 'john', 'a', 'b')

        self.assertEqual(operator, Account(operator.name, wallet_private_key=operator.wallet_private_key))
        self.assertEqual(seller, Account(seller.name, wallet_private_key=seller.wallet_private_key))
        self.assertEqual(buyer, Account(buyer.name, wallet_private_key=buyer.wallet_private_key))

    def test_equals(self):
        self.assertEqual(operator, operator)
        self.assertEqual(seller, seller)
        self.assertEqual(buyer, buyer)

        self.assertEqual(operator, Account(operator.name, operator.wallet_address, operator.wallet_private_key))
        self.assertEqual(seller, Account(seller.name, seller.wallet_address, seller.wallet_private_key))
        self.assertEqual(buyer, Account(buyer.name, buyer.wallet_address, buyer.wallet_private_key))

        self.assertNotEqual(operator, seller)
        self.assertNotEqual(seller, buyer)
        self.assertNotEqual(buyer, operator)



# This file is part of the Blockchain Data Trading Simulator
#    https://gitlab.com/MatthiasLohr/bdtsim
#
# Copyright 2020 Matthias Lohr <mail@mlohr.com>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import argparse
from typing import Dict

from bdtsim.data_provider import DataProviderManager
from bdtsim.environment import EnvironmentManager
from bdtsim.protocol import ProtocolManager
from bdtsim.output import OutputFormatManager
from bdtsim.simulation import Simulation
from .command_manager import SubCommand


class RunSubCommand(SubCommand):
    help = 'run a simulation'

    def __init__(self, parser: argparse.ArgumentParser) -> None:
        super(RunSubCommand, self).__init__(parser)
        parser.add_argument('protocol', choices=ProtocolManager.protocols.keys(), help='protocol to be simulated')
        parser.add_argument('environment', choices=EnvironmentManager.environments.keys(),
                            help='environment in which the simulation will take place')
        parser.add_argument('--data-provider', choices=DataProviderManager.data_providers.keys(),
                            default='RandomDataProvider', help='set the data provider/data source for the simulation')
        parser.add_argument('-f', '--output-format', choices=OutputFormatManager.output_formats.keys(),
                            default='human-readable', help='set the desired output format for simulation results')
        parser.add_argument('--price', type=int, default=1000000000, help='set the price for the asset to be traded')
        parser.add_argument('-p', '--protocol-parameter', nargs=2, action='append', dest='protocol_parameters',
                            default=[], metavar=('KEY', 'VALUE'), help='pass additional parameters to the protocol')
        parser.add_argument('-e', '--environment-parameter', nargs=2, action='append', dest='environment_parameters',
                            default=[], metavar=('KEY', 'VALUE'), help='pass additional parameters to the environment')
        parser.add_argument('-d', '--data-provider-parameter', nargs=2, action='append',
                            dest='data_provider_parameters', default=[], metavar=('KEY', 'VALUE'),
                            help='pass additional parameters to the data provider')
        parser.add_argument('-o', '--output-format-parameter', nargs=2, action='append',
                            dest='output_format_parameters', default=[], metavar=('KEY', 'VALUE'),
                            help='pass additional parameters to the output format')

    def __call__(self, args: argparse.Namespace) -> None:
        protocol_parameters: Dict[str, str] = {}
        environment_parameters: Dict[str, str] = {}
        data_provider_parameters: Dict[str, str] = {}
        output_format_parameters: Dict[str, str] = {}

        for arg, dest in [
            (args.protocol_parameters, protocol_parameters),
            (args.environment_parameters, environment_parameters),
            (args.data_provider_parameters, data_provider_parameters),
            (args.output_format_parameters, output_format_parameters),
        ]:
            for key, value in arg:
                dest[key.replace('-', '_')] = value

        protocol = ProtocolManager.instantiate(
            args.protocol,
            **protocol_parameters
        )

        environment = EnvironmentManager.instantiate(
            args.environment,
            **environment_parameters
        )

        data_provider = DataProviderManager.instantiate(
            args.data_provider,
            **data_provider_parameters
        )

        simulation = Simulation(
            protocol=protocol,
            environment=environment,
            data_provider=data_provider,
            price=args.price
        )

        output_format = OutputFormatManager.instantiate(
            args.output_format,
            **output_format_parameters
        )

        simulation_result = simulation.run()
        output_format.render(simulation_result)



# This file is part of the Blockchain Data Trading Simulator
#    https://gitlab.com/MatthiasLohr/bdtsim
#
# Copyright 2020 Matthias Lohr <mail@mlohr.com>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import itertools
import math
from typing import Any, List, Tuple

from web3 import Web3


class MerkleTreeNode(object):
    def __init__(self, *children: 'MerkleTreeNode') -> None:
        if len(children) > 2:
            raise ValueError('Cannot have more than two children')
        self._children = list(children)

    @property
    def children(self) -> List['MerkleTreeNode']:
        return self._children

    @property
    def leaves(self) -> List['MerkleTreeLeaf']:
        return list(itertools.chain.from_iterable([c.leaves for c in self.children]))

    @property
    def digest(self) -> bytes:
        keccak_inputs: List[Tuple[str, bytes]] = []
        for child in self.children:
            keccak_inputs.append(('bytes32', child.digest))
        types, args = zip(*keccak_inputs)
        return bytes(Web3.solidityKeccak(types, args))

    @property
    def digests_dfs(self) -> List[bytes]:
        return list(itertools.chain.from_iterable([c.digests_dfs for c in self.children])) + [self.digest]

    @property
    def digests_pack(self) -> List[bytes]:
        return [digest for digest, level in sorted(self._digests_pack(0), key=lambda d: d[1], reverse=True)]

    def _digests_pack(self, level: int) -> List[Tuple[bytes, int]]:
        return list(itertools.chain.from_iterable([
            c._digests_pack(level + 1) for c in self.children
        ])) + [(self.digest, level)]

    def has_indirect_child(self, node: 'MerkleTreeNode') -> bool:
        if node in self.children:
            return True

        for child in self.children:
            if child.has_indirect_child(node):
                return True

        return False

    def get_proof(self, node: 'MerkleTreeLeaf') -> List[bytes]:
        if self.children[0] == node:
            return [self.children[1].digest]
        elif self.children[1] == node:
            return [self.children[0].digest]

        if self.children[0].has_indirect_child(node):
            return [self.children[1].digest] + self.children[0].get_proof(node)
        elif self.children[1].has_indirect_child(node):
            return [self.children[0].digest] + self.children[1].get_proof(node)
        else:
            raise ValueError('Node is not part of this tree')

    @staticmethod
    def validate_proof(root_digest: bytes, node: 'MerkleTreeNode', index: int, proof: List[bytes]) -> bool:
        tmp_digest = node.digest
        for i in range(len(proof)):
            if (index & 1 << i) >> i == 1:
                tmp_digest = Web3.solidityKeccak(['bytes32', 'bytes32'], [proof[len(proof) - i - 1], tmp_digest])
            else:
                tmp_digest = Web3.solidityKeccak(['bytes32', 'bytes32'], [tmp_digest, proof[len(proof) - i - 1]])
        return tmp_digest == root_digest

    def __repr__(self) -> str:
        return '<%s.%s %s>' % (
            __name__,
            MerkleTreeNode.__name__,
            self.digest.hex()
        )

    def __eq__(self, other: Any) -> bool:
        if isinstance(other, MerkleTreeNode):
            return self.digest == other.digest
        else:
            return NotImplemented

    def __ne__(self, other: Any) -> bool:
        return not self.__eq__(other)


class MerkleTreeLeaf(MerkleTreeNode):
    def __init__(self, data: bytes) -> None:
        super(MerkleTreeLeaf, self).__init__()
        if (len(data) % 32) != 0:
            raise ValueError('data length has to be a multiple of 32')
        self._data = data

    @property
    def digest(self) -> bytes:
        data_as_list = self.data_as_list()
        # noinspection PyCallByClass
        return bytes(Web3.solidityKeccak(['bytes[%d]' % len(data_as_list)], [data_as_list]))

    @property
    def data(self) -> bytes:
        return self._data

    @data.setter
    def data(self, data: bytes) -> None:
        self._data = data

    def data_as_list(self, slice_size: int = 32) -> List[bytes]:
        return [self.data[i * slice_size:(i + 1) * slice_size] for i in range(int(len(self.data) / slice_size))]

    @property
    def leaves(self) -> List['MerkleTreeLeaf']:
        return [self]

    @property
    def digests_dfs(self) -> List[bytes]:
        return []

    @property
    def digests_pack(self) -> List[bytes]:
        return []

    def _digests_pack(self, level: int) -> List[Tuple[bytes, int]]:
        return []

    def __repr__(self) -> str:
        return '<%s.%s %s>' % (
            __name__,
            MerkleTreeLeaf.__name__,
            str(self.data)
        )

    def __eq__(self, other: Any) -> bool:
        if isinstance(other, MerkleTreeLeaf):
            return self.data == other.data
        else:
            return NotImplemented

    def __ne__(self, other: Any) -> bool:
        return not self.__eq__(other)


class MerkleTreeHashLeaf(MerkleTreeLeaf):
    @property
    def digest(self) -> bytes:
        return self.data


def from_leaves(leaves: List[MerkleTreeLeaf]) -> MerkleTreeNode:
    if len(leaves) == 0:
        raise ValueError('Cannot create tree from empty list')
    nodes: List[MerkleTreeNode] = list(leaves)
    while len(nodes) > 1:
        nodes = [MerkleTreeNode(*nodes[i:i + 2]) for i in range(0, len(nodes), 2)]
    return nodes[0]


def from_bytes(data: bytes, slices_count: int = 8) -> MerkleTreeNode:
    if slices_count < 2 or not math.log2(slices_count).is_integer():
        raise ValueError('slices_count must be >= 2 integer and power of 2')
    slice_len = math.ceil(len(data) / slices_count)
    return from_leaves([MerkleTreeLeaf(data[slice_len * s:slice_len * (s + 1)]) for s in range(slices_count)])


def from_list(items: List[bytes]) -> MerkleTreeNode:
    return from_leaves([MerkleTreeLeaf(item) for item in items])



# This file is part of the Blockchain Data Trading Simulator
#    https://gitlab.com/MatthiasLohr/bdtsim
#
# Copyright 2020 Matthias Lohr <mail@mlohr.com>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import argparse
from typing import Dict

from bdtsim.environment import EnvironmentManager
from .command_manager import SubCommand


class EnvironmentInfoSubCommand(SubCommand):
    help = 'print information about the selected environment'

    def __init__(self, parser: argparse.ArgumentParser) -> None:
        super(EnvironmentInfoSubCommand, self).__init__(parser)
        parser.add_argument('environment', choices=EnvironmentManager.environments.keys())
        parser.add_argument('-e', '--environment-parameter', nargs=2, action='append', dest='environment_parameters',
                            default=[], metavar=('KEY', 'VALUE'), help='Pass additional parameters to the environment')

    def __call__(self, args: argparse.Namespace) -> int:
        environment_parameters: Dict[str, str] = {}
        for key, value in args.environment_parameters:
            environment_parameters[key.replace('-', '_')] = value

        environment = EnvironmentManager.instantiate(
            args.environment,
            **environment_parameters
        )

        print('Client version: %s' % environment.web3.clientVersion)
        syncing = environment.web3.eth.syncing
        if not syncing:
            print('Sync Status: not syncing')
        elif not isinstance(syncing, bool):
            print('Sync Status: syncing (%d/%d)' % (syncing.currentBlock, syncing.highestBlock))

        print('Chain ID: %s' % str(environment.web3.eth.chainId))
        return 0



# This file is part of the Blockchain Data Trading Simulator
#    https://gitlab.com/MatthiasLohr/bdtsim
#
# Copyright 2020 Matthias Lohr <mail@mlohr.com>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import argparse
import logging
from typing import Dict, Optional, Type

import bdtsim


class SubCommand(object):
    help: Optional[str] = None

    def __init__(self, parser: argparse.ArgumentParser):
        self._parser = parser

    def __call__(self, args: argparse.Namespace) -> Optional[int]:
        raise NotImplementedError()


class CommandManager(object):
    def __init__(self) -> None:
        self._argument_parser = argparse.ArgumentParser()
        self._argument_parser.add_argument('-l', '--log-level', default='WARNING',
                                           choices=['DEBUG', 'INFO', 'WARNING', 'ERROR', 'CRITICAL'],
                                           help='set log level for bdtsim')
        self._command_parser = self._argument_parser.add_subparsers(title='command', dest='command', required=True)
        self._subcommands: Dict[str, SubCommand] = {}

    def register_subcommand(self, subcommand: str, cls: Type[SubCommand]) -> None:
        if not issubclass(cls, SubCommand):
            raise ValueError('cls is not a subclass of SubCommand')
        self._subcommands[subcommand] = cls(self._command_parser.add_parser(subcommand, help=cls.help))

    def get_subcommand(self, subcommand: str) -> SubCommand:
        return self._subcommands[subcommand]

    def run(self) -> Optional[int]:
        args = self._argument_parser.parse_args()
        logger = logging.getLogger(bdtsim.__name__)
        logger.setLevel(logging.getLevelName(args.log_level))
        subcommand = self.get_subcommand(args.command)
        return subcommand(args)



# This file is part of the Blockchain Data Trading Simulator
#    https://gitlab.com/MatthiasLohr/bdtsim
#
# Copyright 2020 Matthias Lohr <mail@mlohr.com>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
from typing import Any

from bdtsim.account import Account
from bdtsim.data_provider import DataProvider
from bdtsim.environment import Environment
from bdtsim.protocol_path import ProtocolPath


class Protocol(object):
    def __init__(self, *args: Any, **kwargs: Any) -> None:
        pass

    def prepare_simulation(self, environment: Environment, operator: Account) -> None:
        pass

    def prepare_iteration(self, environment: Environment, operator: Account) -> None:
        pass

    def execute(self, protocol_path: ProtocolPath, environment: Environment, data_provider: DataProvider,
                seller: Account, buyer: Account, price: int = 1000000000) -> None:
        raise NotImplementedError()

    def cleanup_iteration(self, environment: Environment, operator: Account) -> None:
        pass

    def cleanup_simulation(self, environment: Environment, operator: Account) -> None:
        pass

    @staticmethod
    def contract_path(file_var: str, filename: str) -> str:
        return os.path.join(os.path.dirname(file_var), filename)



# This file is part of the Blockchain Data Trading Simulator
#    https://gitlab.com/MatthiasLohr/bdtsim
#
# Copyright 2020 Matthias Lohr <mail@mlohr.com>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
from types import TracebackType
from typing import Any, Dict, Optional, Type

from bdtsim.account import Account
from bdtsim.environment import Environment
from bdtsim.funds_diff_collection import FundsDiffCollection
from bdtsim.protocol_path import ProtocolPath, Decision
from .simulation_result import SimulationResult, ResultNode, TransactionLogEntry, TransactionLogList


logger = logging.getLogger(__name__)


class SimpleTransactionMonitor(object):
    def __init__(self, environment: Environment, transactions_target: TransactionLogList) -> None:
        self._environment = environment
        self._transactions_target = transactions_target

    def _transaction_callback(self, account: Account, tx_dict: Dict[str, Any], tx_receipt: Dict[str, Any],
                              funds_diff_collection: FundsDiffCollection) -> None:
        self._transactions_target.append(TransactionLogEntry(account, tx_dict, tx_receipt, funds_diff_collection))

    def __enter__(self) -> None:
        self._environment.transaction_callback = self._transaction_callback

    def __exit__(self, exception_type: Optional[Type[BaseException]], exception: Optional[BaseException],
                 traceback: Optional[TracebackType]) -> None:
        self._environment.transaction_callback = None


class ExecutionTransactionMonitor(object):
    def __init__(self, environment: Environment, protocol_path: ProtocolPath,
                 execution_result_root: ResultNode) -> None:
        self._environment = environment
        self._protocol_path = protocol_path
        self._execution_result_root = execution_result_root

        self._current_execution_result_node: Optional[ResultNode] = None
        self._current_transactions: TransactionLogList = TransactionLogList()

    def _decision_callback(self, decision: Decision) -> None:
        if self._current_execution_result_node is None:
            raise RuntimeError()
        self._current_execution_result_node.tx_collection.append(self._current_transactions)
        self._current_transactions = TransactionLogList()
        self._current_execution_result_node = self._current_execution_result_node.child(decision)

    def _transaction_callback(self, account: Account, tx_dict: Dict[str, Any], tx_receipt: Dict[str, Any],
                              funds_diff_collection: FundsDiffCollection) -> None:
        self._current_transactions.append(TransactionLogEntry(account, tx_dict, tx_receipt, funds_diff_collection))

    def __enter__(self) -> None:
        self._current_execution_result_node = self._execution_result_root
        self._environment.transaction_callback = self._transaction_callback
        self._protocol_path.decision_callback = self._decision_callback

    def __exit__(self, exception_type: Optional[Type[BaseException]], exception: Optional[BaseException],
                 traceback: Optional[TracebackType]) -> None:
        if self._current_execution_result_node is None:
            raise RuntimeError()
        self._current_execution_result_node.tx_collection.append(self._current_transactions)
        self._protocol_path.decision_callback = None
        self._environment.transaction_callback = None


class ResultCollector(object):
    def __init__(self) -> None:
        self.simulation_result = SimulationResult()

    def monitor_preparation(self, environment: Environment) -> SimpleTransactionMonitor:
        return SimpleTransactionMonitor(environment, self.simulation_result.preparation_transactions)

    def monitor_execution(self, environment: Environment, protocol_path: ProtocolPath) -> ExecutionTransactionMonitor:
        return ExecutionTransactionMonitor(environment, protocol_path, self.simulation_result.execution_result_root)

    def monitor_cleanup(self, environment: Environment) -> SimpleTransactionMonitor:
        return SimpleTransactionMonitor(environment, self.simulation_result.cleanup_transactions)



# This file is part of the Blockchain Data Trading Simulator
#    https://gitlab.com/MatthiasLohr/bdtsim
#
# Copyright 2020 Matthias Lohr <mail@mlohr.com>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import unittest

from bdtsim.util.types import to_bool


class TypeUtilTest(unittest.TestCase):
    def test_to_bool(self):
        self.assertEqual(True, to_bool(True))
        self.assertEqual(True, to_bool('True'))
        self.assertEqual(True, to_bool('true'))
        self.assertEqual(True, to_bool('Yes'))
        self.assertEqual(True, to_bool('yes'))
        self.assertEqual(True, to_bool('Y'))
        self.assertEqual(True, to_bool('y'))
        self.assertEqual(True, to_bool('1'))

        self.assertEqual(False, to_bool(False))
        self.assertEqual(False, to_bool('False'))
        self.assertEqual(False, to_bool('false'))
        self.assertEqual(False, to_bool('No'))
        self.assertEqual(False, to_bool('no'))
        self.assertEqual(False, to_bool('N'))
        self.assertEqual(False, to_bool('n'))
        self.assertEqual(False, to_bool('0'))

        self.assertRaises(ValueError, to_bool, 'foobar')



# This file is part of the Blockchain Data Trading Simulator
#    https://gitlab.com/MatthiasLohr/bdtsim
#
# Copyright 2020 Matthias Lohr <mail@mlohr.com>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging

from bdtsim.contract import SolidityContract
from bdtsim.data_provider import DataProvider
from bdtsim.environment import Environment
from bdtsim.account import Account
from bdtsim.protocol import Protocol, ProtocolManager
from bdtsim.protocol_path import ProtocolPath


logger = logging.getLogger(__name__)


class SimplePayment(Protocol):
    """Example implementation of SimplePayment (paying after exchange, direct payments)"""
    def execute(self, protocol_path: ProtocolPath, environment: Environment, data_provider: DataProvider,
                seller: Account, buyer: Account, price: int = 1000000000) -> None:
        release_goods = protocol_path.decide(seller, 'release goods?', ['yes', 'no'])
        # [...] seller releases goods to buyer, not monitored by BDTsim
        if release_goods == 'yes':
            pay = protocol_path.decide(buyer, 'pay?', ['yes', 'no'])
            if pay == 'yes':
                environment.send_direct_transaction(buyer, seller, price)
            else:
                pass  # buyer left protocol, having goods, without payment
        else:
            pass  # seller left protocol, no reaction from buyer


class AbstractParameterizedSimplePayment(Protocol):
    def __init__(self, use_contract: bool):
        super(AbstractParameterizedSimplePayment, self).__init__()
        self._use_contract = use_contract

    def get_contract(self) -> SolidityContract:
        return SolidityContract('SimplePayment', self.contract_path(__file__, 'SimplePayment.sol'))

    def prepare_simulation(self, environment: Environment, operator: Account) -> None:
        if self._use_contract:
            logger.debug('Deploying contract...')
            environment.deploy_contract(operator, self.get_contract())

    def execute(self, protocol_path: ProtocolPath, environment: Environment, data_provider: DataProvider,
                seller: Account, buyer: Account, price: int = 1000000000) -> None:
        raise NotImplementedError()


class SimplePaymentPrepaid(AbstractParameterizedSimplePayment):
    def execute(self, protocol_path: ProtocolPath, environment: Environment, data_provider: DataProvider,
                seller: Account, buyer: Account, price: int = 1000000000) -> None:
        if protocol_path.decide(buyer, description='Payment', variants=['paying', 'not paying']).is_honest():
            logger.debug('Decided to be honest')
            if self._use_contract:
                environment.send_contract_transaction(buyer, 'pay', seller.wallet_address, value=price)
            else:
                environment.send_direct_transaction(buyer, seller, price)

            protocol_path.decide(seller, description='Give goods', variants=['yes', 'no'])
        else:
            logger.debug('Decided to cheat')  # do nothing


class SimplePaymentPostpaid(AbstractParameterizedSimplePayment):
    def execute(self, protocol_path: ProtocolPath, environment: Environment, data_provider: DataProvider,
                seller: Account, buyer: Account, price: int = 1000000000) -> None:
        if protocol_path.decide(seller, 'Give goods', ['yes', 'no']).is_honest():
            if protocol_path.decide(buyer, description='Payment', variants=['paying', 'not paying']).is_honest():
                logger.debug('Decided to be honest')
                if self._use_contract:
                    environment.send_contract_transaction(buyer, 'pay', seller.wallet_address, value=price)
                else:
                    environment.send_direct_transaction(buyer, seller, price)


ProtocolManager.register('SimplePayment', SimplePayment)
ProtocolManager.register('SimplePayment-prepaid', SimplePaymentPrepaid, use_contract=True)
ProtocolManager.register('SimplePayment-prepaid-direct', SimplePaymentPrepaid, use_contract=False)
ProtocolManager.register('SimplePayment-postpaid', SimplePaymentPostpaid, use_contract=True)
ProtocolManager.register('SimplePayment-postpaid-direct', SimplePaymentPostpaid, use_contract=False)



# This file is part of the Blockchain Data Trading Simulator
#    https://gitlab.com/MatthiasLohr/bdtsim
#
# Copyright 2020 Matthias Lohr <mail@mlohr.com>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import math
import unittest

from eth_tester import EthereumTester, PyEVMBackend
from eth_utils.crypto import keccak
from web3 import Web3, EthereumTesterProvider

from bdtsim.account import buyer
from bdtsim.protocol.fairswap import FairSwap
from bdtsim.protocol.fairswap.encoding import encode, decode, B032, crypt
from bdtsim.protocol.fairswap.merkle import MerkleTreeNode, MerkleTreeLeaf, from_bytes


class FairSwapTest(unittest.TestCase):
    def test_keccak(self):
        data = FairSwap.generate_bytes(32)
        self.assertEqual(keccak(data), Web3.solidityKeccak(['bytes32'], [data]))
        self.assertEqual(keccak(data), Web3.solidityKeccak(['bytes32[1]'], [[data]]))
        self.assertEqual(keccak(data), Web3.keccak(data))


class MerkleTest(unittest.TestCase):
    def test_init(self):
        self.assertRaises(ValueError, MerkleTreeNode, MerkleTreeLeaf(B032), MerkleTreeLeaf(B032), MerkleTreeLeaf(B032))

    def test_get_proof_and_validate(self):
        for slice_count in [2, 4, 8, 16]:
            tree = from_bytes(FairSwap.generate_bytes(32 * slice_count), slice_count)
            for index, leaf in enumerate(tree.leaves):
                proof = tree.get_proof(leaf)
                self.assertEqual(len(proof), int(math.log2(slice_count)))
                self.assertTrue(MerkleTreeNode.validate_proof(tree.digest, leaf, index, proof))


class EncodingTest(unittest.TestCase):
    def test_encode_decode(self):
        tree = from_bytes(FairSwap.generate_bytes(128, seed=42), 4)
        tree_enc = encode(tree, B032)
        tree_dec, errors = decode(tree_enc, B032)
        self.assertEqual([], errors)
        self.assertEqual(tree, tree_dec)


class ContractTest(unittest.TestCase):
    @staticmethod
    def prepare_contract(file_root_hash, ciphertext_root_hash, key_hash, slice_count: int = 4):
        web3 = Web3(EthereumTesterProvider(EthereumTester(PyEVMBackend())))
        contract_object = FairSwap(slice_count)._get_contract(
            buyer=buyer,
            price=1000000000,
            slice_length=32,
            file_root_hash=file_root_hash,
            ciphertext_root_hash=ciphertext_root_hash,
            key_hash=key_hash
        )
        contract_preparation = web3.eth.contract(abi=contract_object.abi, bytecode=contract_object.bytecode)
        tx_hash = contract_preparation.constructor().transact()
        tx_receipt = web3.eth.waitForTransactionReceipt(tx_hash)
        return web3, web3.eth.contract(address=tx_receipt.contractAddress, abi=contract_object.abi)

    def test_vrfy(self):
        tree = from_bytes(FairSwap.generate_bytes(128, seed=42), 4)
        key = FairSwap.generate_bytes(32, seed=43)
        tree_enc = encode(tree, key)
        web3, contract = self.prepare_contract(tree.digest, tree_enc.digest, B032)

        for index, leaf in enumerate(tree_enc.leaves):
            proof = tree_enc.get_proof(leaf)
            self.assertEqual(len(proof), 3)
            call_result = contract.functions.vrfy(index, leaf.digest, proof).call()
            self.assertTrue(call_result)

    def test_crypt_small(self):
        for n in [4, 8, 16]:
            web3, contract = self.prepare_contract(FairSwap.generate_bytes(32), FairSwap.generate_bytes(32), B032, n)
            for i in range(8):
                data = FairSwap.generate_bytes(32)
                call_result = contract.functions.cryptSmall(i, data).call()
                self.assertEqual(call_result, crypt(data, n + i, B032))



# This file is part of the Blockchain Data Trading Simulator
#    https://gitlab.com/MatthiasLohr/bdtsim
#
# Copyright 2020 Matthias Lohr <mail@mlohr.com>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Any


def to_bool(obj: Any) -> bool:
    if isinstance(obj, bool):
        return obj
    str_value = str(obj).lower()
    if str_value in ['true', 'yes', 'y', '1']:
        return True
    elif str_value in ['false', 'no', 'n', '0']:
        return False
    else:
        raise ValueError('%s can not be converted to bool' % str_value)



# This file is part of the Blockchain Data Trading Simulator
#    https://gitlab.com/MatthiasLohr/bdtsim
#
# Copyright 2020 Matthias Lohr <mail@mlohr.com>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import math
from typing import Any, Optional


class FileSize(object):
    SUFFIXES = ['k', 'M', 'G', 'T', 'P', 'E', 'Z', 'Y']

    def __init__(self, value: Any) -> None:
        self._size = self.parse_value(value)

    @staticmethod
    def parse_value(value: Any) -> int:
        if isinstance(value, int):
            return value
        if isinstance(value, str):
            suffix = value[-1:].lower()
            suffix_index = None
            for i, s in enumerate(FileSize.SUFFIXES):
                if s.lower() == suffix:
                    suffix_index = i
                    break

            if suffix_index is None:
                return int(value)
            else:
                return int(float(value[:-1]) * (1000 ** (suffix_index + 1)))
        else:
            raise ValueError('Unsupported type: %s' % type(value))

    @staticmethod
    def format_human_readable(b: int, ndigits: Optional[int] = None) -> str:
        log = math.log(b, 1000)
        if log >= 1:
            exp = math.floor(log)
            return str(round(b / 1000 ** exp, ndigits)) + FileSize.SUFFIXES[exp - 1]
        else:
            return str(b)

    def human_readable(self) -> str:
        return self.format_human_readable(self._size)

    def __eq__(self, other: Any) -> bool:
        if isinstance(other, FileSize):
            return self._size == other._size
        elif isinstance(other, int):
            return self._size == other
        elif isinstance(other, str):
            return self._size == FileSize(other)._size
        else:
            raise NotImplementedError()

    def __ne__(self, other: Any) -> bool:
        return not self.__eq__(other)

    def __repr__(self) -> str:
        return '<%s.%s %d bytes (%s)>' % (
            __name__,
            FileSize.__name__,
            self._size,
            self.format_human_readable(self._size)
        )

    def __str__(self) -> str:
        return '<%s %d bytes (%s)>' % (
            FileSize.__name__,
            self._size,
            self.format_human_readable(self._size)
        )



# This file is part of the Blockchain Data Trading Simulator
#    https://gitlab.com/MatthiasLohr/bdtsim
#
# Copyright 2020 Matthias Lohr <mail@mlohr.com>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Any, Optional

from eth_typing.evm import ChecksumAddress
from web3 import Web3


class Account(object):
    def __init__(self, name: str, wallet_address: Optional[str] = None, wallet_private_key: Optional[str] = None)\
            -> None:
        if wallet_address is None and wallet_private_key is None:
            raise ValueError('You have to provide at least wallet_address OR wallet_private_key')
        elif wallet_address is not None and wallet_private_key is not None:
            eth_account = Web3().eth.account.from_key(wallet_private_key)
            if eth_account.address != wallet_address:
                raise ValueError('wallet_private_key does not match wallet_address')
        elif wallet_address is None and wallet_private_key is not None:
            wallet_address = Web3().eth.account.from_key(wallet_private_key).address
        self._name = name
        self._wallet_address: ChecksumAddress = Web3.toChecksumAddress(wallet_address)
        self._wallet_private_key = wallet_private_key

    @property
    def name(self) -> str:
        return self._name

    @property
    def wallet_address(self) -> ChecksumAddress:
        return self._wallet_address

    @property
    def wallet_private_key(self) -> Optional[str]:
        return self._wallet_private_key

    def __str__(self) -> str:
        return '%s<%s>' % (self.name, self.wallet_address)

    def __repr__(self) -> str:
        return '<%s.%s name=%s wallet_address=%s has_wallet_private_key=%s>' % (
            __name__,
            self.__class__.__name__,
            self.name,
            self.wallet_address,
            str(self.wallet_private_key is not None)
        )

    def __eq__(self, other: Any) -> bool:
        if isinstance(other, Account):
            return self.wallet_address == other.wallet_address and self.name == other.name
        else:
            raise NotImplementedError()

    def __ne__(self, other: Any) -> bool:
        return not self.__eq__(other)

    def __hash__(self) -> int:
        return hash((self.wallet_address, self.name))


operator = Account(
    name='Operator',
    wallet_address='0x3ed8424aaE568b3f374e94a139D755982800a4a2',
    wallet_private_key='0xe67518b4d5255ec708d2bf9cd4222adda89fcc07037c614d7787a694fbb47692'
)

seller = Account(
    name='Seller',
    wallet_address='0x5Afa5874959ff249103c2043fB45d68B2768Fef8',
    wallet_private_key='0x3df2d74ceb3c58a8fdb1f0ecf45e2ceb10511469d9c20691333d666fa557899a'
)

buyer = Account(
    name='Buyer',
    wallet_address='0x00c382926f098566EA6F1707296eC342E7C8A5DC',
    wallet_private_key='0x7d96e8fbe712cf25f141adb6bc5e3244d7a19d9c406ab6ed6a097585d01b93ac'
)



# This file is part of the Blockchain Data Trading Simulator
#    https://gitlab.com/MatthiasLohr/bdtsim
#
# Copyright 2020 Matthias Lohr <mail@mlohr.com>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import random
from io import BytesIO
from typing import BinaryIO, Optional

from .data_provider import DataProvider
from .data_provider_manager import DataProviderManager


class RandomDataProvider(DataProvider):
    def __init__(self, size: int = 1048576, seed: int = 42) -> None:
        super(RandomDataProvider, self).__init__()
        self._size = int(size)
        self._seed = seed
        self._mem_file: Optional[BytesIO] = None

    @property
    def data_size(self) -> int:
        return self._size

    @property
    def file_pointer(self) -> BinaryIO:
        if self._mem_file is None:
            random.seed(self._seed)
            self._mem_file = BytesIO(bytearray(random.getrandbits(8) for _ in range(self._size)))
        return self._mem_file


DataProviderManager.register('RandomDataProvider', RandomDataProvider)



# This file is part of the Blockchain Data Trading Simulator
#    https://gitlab.com/MatthiasLohr/bdtsim
#
# Copyright 2020 Matthias Lohr <mail@mlohr.com>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
from queue import Queue

from bdtsim.account import buyer, operator, seller
from bdtsim.data_provider import DataProvider
from bdtsim.environment import Environment
from bdtsim.protocol import Protocol
from bdtsim.protocol_path import ProtocolPath
from bdtsim.output import SimulationResult, ResultCollector


logger = logging.getLogger(__name__)


class Simulation(object):
    def __init__(self, protocol: Protocol, environment: Environment, data_provider: DataProvider,
                 price: int = 1000000000) -> None:
        self._protocol = protocol
        self._environment = environment
        self._data_provider = data_provider
        self._price = price

        self._protocol_path_queue: Queue[ProtocolPath] = Queue()

    def run(self) -> SimulationResult:
        result_collector = ResultCollector()

        logger.debug('Preparing environment for simulation...')
        with result_collector.monitor_preparation(self._environment):
            self._protocol.prepare_simulation(self._environment, operator)
        logger.debug('Finished preparing the environment for simulation')

        self._protocol_path_queue.put(ProtocolPath())
        logger.debug('Starting simulation loop')
        while not self._protocol_path_queue.empty():
            protocol_path = self._protocol_path_queue.get(block=False)

            with result_collector.monitor_execution(self._environment, protocol_path):
                logger.debug('Simulation will follow path %s' % str(protocol_path))

                logger.debug('Preparing environment for iteration...')
                self._protocol.prepare_iteration(self._environment, operator)
                logger.debug('Finished preparing the environment for iteration')

                self._data_provider.file_pointer.seek(0, 0)
                logger.debug('Starting protocol execution...')
                self._protocol.execute(
                    protocol_path=protocol_path,
                    environment=self._environment,
                    data_provider=self._data_provider,
                    seller=seller,
                    buyer=buyer,
                    price=self._price
                )
                logger.debug('Finished protocol execution')

                logger.debug('Starting cleanup for iteration...')
                self._protocol.cleanup_iteration(self._environment, operator)
                logger.debug('Finished cleaning up the iteration')

                logger.debug('Collecting alternative paths...')

                for alternative_path in protocol_path.get_alternatives():
                    self._protocol_path_queue.put(alternative_path)
                    logger.debug('Added new path %s' % str(alternative_path))

            self._protocol_path_queue.task_done()

        logger.debug('Simulation finished. Cleaning up...')
        self._protocol.cleanup_simulation(self._environment, operator)
        logger.debug('Finished cleaning up the simulation')
        return result_collector.simulation_result



# This file is part of the Blockchain Data Trading Simulator
#    https://gitlab.com/MatthiasLohr/bdtsim
#
# Copyright 2020 Matthias Lohr <mail@mlohr.com>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import time
from typing import Any, Callable, List, Optional

from .account import Account


class Decision(object):
    """A single decision within a `ProtocolPath`.

    Decisions are represented by integers (starting with 1).
    By convention, 1 means "honest"/"most honest" decision, all other values (2, 3, ...) represent a dishonest/cheating
    decision.
    """

    def __init__(self, account: Account, outcome: str, variants: List[str], honest_variants: Optional[List[str]] = None,
                 description: Optional[str] = None, timestamp: Optional[float] = None) -> None:
        """
        Args:
            account (Account): Who is deciding about the next step
            outcome (str): Variant chosen in this decision
            variants (List[str]): Possible variants
            honest_variants (List[str]): List of variants to be considered as honest
            description (str): Description of this decision (not considered for equality)
            timestamp (float): When the decision has taken place
        """
        if len(variants) < 1:
            raise ValueError('You have to provide at least one variant for a decision or two to have a choice')
        if outcome not in variants:
            raise ValueError('%s not in list of possible variants (%s)' % (outcome, ', '.join(variants)))
        if honest_variants is not None:
            for honest_variant in honest_variants:
                if honest_variant not in variants:
                    raise ValueError('Honest variants have to be part of possible variants list')

        self._account = account
        self._outcome = outcome
        self._variants = variants
        self._honest_variants: List[str] = honest_variants or [variants[0]]
        self._timestamp = timestamp
        self._description = description

    @property
    def account(self) -> Account:
        return self._account

    @property
    def outcome(self) -> str:
        return self._outcome

    @property
    def variants(self) -> List[str]:
        return self._variants

    @property
    def honest_variants(self) -> List[str]:
        return self._honest_variants

    @property
    def timestamp(self) -> Optional[float]:
        return self._timestamp

    @timestamp.setter
    def timestamp(self, timestamp: float) -> None:
        self._timestamp = timestamp

    @property
    def description(self) -> Optional[str]:
        return self._description

    def is_honest(self) -> bool:
        return self.outcome in self._honest_variants

    def is_outcome(self, variant: str) -> bool:
        return self.outcome == variant

    def __eq__(self, other: Any) -> bool:
        """Supports equality checks with other Decision instances and int(egers).

        When an string is provided, the string value is compared to the `outcome` field.

        Args:
            other: (Decision, str): Object to compare to

        Returns:
            bool: Equality
        """
        if isinstance(other, Decision):
            return (self.account == other.account
                    and self.outcome == other.outcome
                    and self.variants == other.variants
                    and self._honest_variants == other._honest_variants)
        elif isinstance(other, str):
            if other not in self.variants:
                raise ValueError('%s is not an allowed variant (%s)' % (other, ', '.join(self.variants)))
            return other == self.outcome
        else:
            raise NotImplementedError()

    def __ne__(self, other: Any) -> bool:
        return not self.__eq__(other)

    def __hash__(self) -> int:
        return hash((self._account, self._outcome, tuple(self._variants)))

    def __repr__(self) -> str:
        return '<%s.%s by %s: %s (%s)>' % (
            __name__,
            Decision.__name__,
            self.account.name,
            self.outcome,
            ', '.join(self.variants)
        )

    def __str__(self) -> str:
        return '%s: %s (%s)' % (
            self.account.name,
            self.outcome,
            ', '.join(self.variants)
        )


class ProtocolPath(object):
    """One possible path through a protocol iteration."""

    def __init__(self, initial_decisions: Optional[List[Decision]] = None) -> None:
        self._initial_decisions: List[Decision] = initial_decisions or []
        self._new_decisions: List[Decision] = []
        self._decisions_index: int = 0
        self._decision_callback: Optional[Callable[[Decision], None]] = None

    def decide(self, account: Account, description: str, variants: List[str],
               honest_variants: Optional[List[str]] = None) -> Decision:
        if len(self.decisions) == self._decisions_index:
            # we have no decision yet, creating a new one
            self._new_decisions.append(Decision(
                account=account,
                outcome=variants[0],
                variants=variants,
                honest_variants=honest_variants,
                timestamp=time.time(),
                description=description
            ))

        decision = self.decisions[self._decisions_index]

        # if there is a pre-defined decision, set timestamp when it was used (now)
        if decision.timestamp is None:
            decision.timestamp = time.time()

        # do some decision validity checks
        if decision.account != account:
            raise ValueError('Account provided does not match pre-defined decision account %s' % str(decision.account))
        if decision.variants != variants:
            raise ValueError('Possible variants do not match')

        self._decisions_index += 1
        if self._decision_callback is not None:
            self._decision_callback(decision)
        return decision

    @property
    def initial_decisions(self) -> List[Decision]:
        return self._initial_decisions

    @property
    def new_decisions(self) -> List[Decision]:
        return self._new_decisions

    @property
    def decisions(self) -> List[Decision]:
        return self.initial_decisions + self.new_decisions

    @property
    def decision_callback(self) -> Optional[Callable[[Decision], None]]:
        return self._decision_callback

    @decision_callback.setter
    def decision_callback(self, callback: Optional[Callable[[Decision], None]]) -> None:
        self._decision_callback = callback

    def get_alternatives(self) -> List['ProtocolPath']:
        alternatives = []
        for new_decision_index in range(len(self.new_decisions)):
            decision_head = self.new_decisions[new_decision_index]
            for variant in decision_head.variants:
                if decision_head.outcome == variant:
                    continue
                alternatives.append(ProtocolPath(
                    self.initial_decisions + self.new_decisions[:new_decision_index] + [Decision(
                        account=decision_head.account,
                        outcome=variant,
                        variants=decision_head.variants,
                        honest_variants=decision_head._honest_variants,
                        timestamp=None,
                        description=decision_head.description
                    )]
                ))
        return alternatives

    def all_accounts_completely_honest(self) -> bool:
        for decision in self.decisions:
            if not decision.is_honest():
                return False
        return True

    def account_completely_honest(self, account: Account) -> bool:
        for decision in self.decisions:
            if decision.account != account:
                continue
            if not decision.is_honest():
                return False
        return True

    def __eq__(self, other: Any) -> bool:
        if isinstance(other, ProtocolPath):
            return self.initial_decisions == other.initial_decisions and self.new_decisions == other.new_decisions
        else:
            raise NotImplementedError()

    def __ne__(self, other: Any) -> bool:
        return not self.__eq__(other)

    def __repr__(self) -> str:
        initial_decisions_str = map(lambda d: str(d), self.initial_decisions)
        new_decisions_str = map(lambda d: str(d), self.new_decisions)
        return '<%s.%s: [%s]+[%s]>' % (
            __name__,
            ProtocolPath.__name__,
            ', '.join(initial_decisions_str),
            ', '.join(new_decisions_str),
        )



# This file is part of the Blockchain Data Trading Simulator
#    https://gitlab.com/MatthiasLohr/bdtsim
#
# Copyright 2020 Matthias Lohr <mail@mlohr.com>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from .output_format import OutputFormat
from .output_format_manager import OutputFormatManager
from .simulation_result import SimulationResult


class HumanReadableOutputFormat(OutputFormat):
    def __init__(self) -> None:
        super(HumanReadableOutputFormat, self).__init__()

    def render(self, simulation_result: SimulationResult) -> None:
        if len(simulation_result.preparation_transactions.aggregation.entries) > 0:
            print('Preparation Results:')
            for preparation_entry in simulation_result.preparation_transactions.aggregation.entries.values():
                print('  %s' % str(preparation_entry))

        print('Execution Results:')
        for description, aggregation in simulation_result.get_important_execution_results():
            print('  %s' % description)
            for execution_entry in aggregation.entries.values():
                print('    %s' % str(execution_entry))

        if len(simulation_result.cleanup_transactions.aggregation.entries) > 0:
            print('Cleanup Results:')
            for cleanup_entry in simulation_result.cleanup_transactions.aggregation.entries.values():
                print('  %s' % cleanup_entry)


OutputFormatManager.register('human-readable', HumanReadableOutputFormat)



# This file is part of the Blockchain Data Trading Simulator
#    https://gitlab.com/MatthiasLohr/bdtsim
#
# Copyright 2020 Matthias Lohr <mail@mlohr.com>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import unittest

from bdtsim.data_provider import RandomDataProvider


class RandomDataProviderTest(unittest.TestCase):
    def test_random_data_provider_size(self):
        for size in 0, 1, 13, 42, 1337, 1000000:
            data_provider = RandomDataProvider(size)
            fp = data_provider.file_pointer
            fp.seek(0, 2)
            self.assertEqual(size, fp.tell())

    def test_random_data_provider_seed(self):
        data = RandomDataProvider(size=4, seed=42).file_pointer.read()
        self.assertEqual(b'\xa3\x1c\x06\xbd', data)
        data = RandomDataProvider(size=4, seed=42).file_pointer.read()
        self.assertEqual(b'\xa3\x1c\x06\xbd', data)
        data = RandomDataProvider(size=4, seed=43).file_pointer.read()
        self.assertEqual(b'\tI\xb2\xc3', data)
        data = RandomDataProvider(size=4, seed=42).file_pointer.read()
        self.assertEqual(b'\xa3\x1c\x06\xbd', data)
